# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

import time
from pathlib import Path

import numpy as np
import pytest
from pyproj import CRS, network

from weitsicht.geometry.intersection_plane import intersection_plane_mat_operation
from weitsicht.mapping.raster import MappingRaster

network.set_network_enabled(True)  # type: ignore

FIXTURE_DIR = Path(__file__).parent.resolve() / "data"


@pytest.fixture
def raster_mapper():
    mapper_raster = MappingRaster(
        raster_path=FIXTURE_DIR / "dhm_at_lamb_10m_2018.tif",
        # raster_path=r"E:\03_australia\dhm_at_lamb_10m_2018.tif",
        crs=CRS("EPSG:31287+5778"),
    )
    limit = (
        551842.0000000000000000,
        426123.0000000000000000,
        561956.0000000000000000,
        433466.0000000000000000,
    )
    mapper_raster.load_window(limit)

    return mapper_raster


@pytest.fixture
def raster_mapper_full():
    mapper_raster = MappingRaster(
        raster_path=FIXTURE_DIR / "dhm_at_lamb_10m_2018.tif",
        # raster_path=r"E:\03_australia\dhm_at_lamb_10m_2018.tif",
        crs=CRS("EPSG:31287+5778"),
        preload_full_raster=True,
    )

    return mapper_raster


def test_map_coordinates_from_rays(raster_mapper):
    ray_crs = CRS("EPSG:31287+5778")  # CRS("EPSG:25833+5778")

    point_3d = np.array(
        [
            [557057, 429465.0, 1520 + 200],
            [557057, 429465.0, 1520 + 200],
            [557057 + 20, 429465.0 + 20, 1520 + 200],
            [557057 + 20, 429465.0 + 20, 1520 + 200],
            [551842.0 - 10, 433466.0 - 50, 1520 + 200],
            [
                551842.0 - 200,
                433466.0 - 50,
                1520 + 200,
            ],  # That point is starting from outside the raster
            [551842.0 - 10, 433466.0 - 50, 1200],
            [551842.0 + 10, 433466.0 - 50, 1130],
            [551842.0 - 10, 433466.0 - 50, 1130],
            [551842.0 + 10, 433466.0, 1720],
        ]
    )

    ray_vec = np.array(
        [
            [0.3, -0.4, -0.80],
            [0.1, -0.3, -1.0],
            [0.3, -0.4, -0.80],
            [0.1, -0.3, -1.0],
            [0.4, 0.0, -1.0],
            [0.4, 0.0, -1.0],
            [0.6, 0.0, -0.8],
            [0.0, -1.0, -0.001],
            [1.0, 0.0, -0.01],
            [0.0, -4.0, -1],
        ]
    )

    b = raster_mapper.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs
    )
    assert b.ok is True
    # We will test 2 things
    # (1) if the intersection point is correct
    # (2) if the bilinear interpolation of that coordinate is correct
    # (3) if that is correct than the intersection is for sure correct as this is on the line
    # and the bilinear interpolated height for that point is also correct
    interp_plane, valid_mas = intersection_plane_mat_operation(ray_vec, point_3d, b.coordinates, np.array([0, 0, 1]))
    assert np.allclose(interp_plane, b.coordinates, atol=1e-7, rtol=0)
    bilin_coo = raster_mapper.georef_array.map_heights_from_coordinates(b.coordinates[:, :2], raster_mapper.crs)
    assert bilin_coo.ok is True
    assert np.allclose(b.coordinates, bilin_coo.coordinates, atol=1e-7, rtol=0)

    # This will test different directions of rays and if the line grid cell intersection delivers all cells
    ray_vec = np.array(
        [
            [0.3, 0.1, -0.80],
            [-0.3, 0.1, -0.80],
            [-0.3, -0.1, -0.80],
            [0.3, -0.1, -0.80],
            [0.3, 0.4, -0.80],
        ]
    )
    point_3d = np.zeros(ray_vec.shape) + np.array([557057, 429465.0, 1520 + 200])

    b = raster_mapper.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs
    )
    assert b.ok is True
    interp_plane, valid_mas = intersection_plane_mat_operation(ray_vec, point_3d, b.coordinates, np.array([0, 0, 1]))
    assert np.allclose(interp_plane, b.coordinates, atol=1e-7, rtol=0)
    bilin_coo = raster_mapper.georef_array.map_heights_from_coordinates(b.coordinates[:, :2], raster_mapper.crs)
    assert np.allclose(b.coordinates, bilin_coo.coordinates, atol=1e-7, rtol=0)

    ray_vec = np.array(
        [
            [-0.3, -0.1, -0.80],
            [-0.3, -0.12, -0.80],
            [-0.3, -0.1, -0.80],
            [-0.3, -0.1, -0.80],
            [-0.4, -0.4, -0.80],  # That one is special for testing
        ]
    )
    point_3d = np.zeros(ray_vec.shape) + np.array([557057, 429465.0, 1520 + 200])

    b = raster_mapper.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs
    )
    interp_plane, valid_mas = intersection_plane_mat_operation(ray_vec, point_3d, b.coordinates, np.array([0, 0, 1]))
    assert np.allclose(interp_plane, b.coordinates, atol=1e-7, rtol=0)
    bilin_coo = raster_mapper.georef_array.map_heights_from_coordinates(b.coordinates[:, :2], raster_mapper.crs)
    assert np.allclose(b.coordinates, bilin_coo.coordinates, atol=1e-7, rtol=0)

    # horizontal rays on X
    point_3d = np.array([[551842.0 - 10, 433466.0 - 50, 1130], [551842.0 - 10, 433466.0 - 50, 1130]])
    # This gives an error -> Check
    ray_vec = np.array([[1.0, -0.1, 0.0], [1.0, 0.0, 0]])
    b = raster_mapper.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs
    )
    # That plane needs to be vertical, otherwise that horizontal ray is parallel
    interp_plane, valid_mas = intersection_plane_mat_operation(
        ray_vec, point_3d, b.coordinates, np.array([0.1, 0.0, 0.0])
    )
    assert np.allclose(interp_plane, b.coordinates, atol=1e-7, rtol=0)
    bilin_coo = raster_mapper.georef_array.map_heights_from_coordinates(b.coordinates[:, :2], raster_mapper.crs)
    assert np.allclose(b.coordinates, bilin_coo.coordinates, atol=1e-7, rtol=0)

    # horizontal rays on Y
    point_3d = np.array([[551842.0 + 10, 433466.0 - 50, 1130], [551842.0 + 10, 433466.0 - 50, 1130]])
    # This gives an error -> Check
    ray_vec = np.array([[0.0, -1.0, 0.0], [0.0, -1.0, 0]])
    b = raster_mapper.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs
    )

    # That plane needs to be vertical, otherwise that horizontal ray is parallel
    interp_plane, valid_mas = intersection_plane_mat_operation(ray_vec, point_3d, b.coordinates, np.array([0, -1.0, 0.0]))
    assert np.allclose(interp_plane, b.coordinates, atol=1e-7, rtol=0)
    bilin_coo = raster_mapper.georef_array.map_heights_from_coordinates(b.coordinates[:, :2], raster_mapper.crs)
    assert np.allclose(b.coordinates, bilin_coo.coordinates, atol=1e-7, rtol=0)


def test_speed(raster_mapper_full):
    ray_crs = CRS("EPSG:31287+5778")  # CRS("EPSG:25833+5778")
    a = time.time()
    ray_vec = np.random.random((1000, 3)) / 100.0 + np.array([0, 0, -1])
    point_3d = np.random.random((1000, 3)) / 10.0 + np.array([556782, 429523, 1600])

    b = raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs
    )
    print(time.time() - a)
    interp_plane, valid_mas = intersection_plane_mat_operation(ray_vec, point_3d, b.coordinates)
    assert np.allclose(interp_plane, b.coordinates, atol=1e-7, rtol=0)
    bilin_coo = raster_mapper_full.georef_array.map_heights_from_coordinates(b.coordinates[:, :2], ray_crs)

    assert np.allclose(b.coordinates, bilin_coo.coordinates, atol=1e-7, rtol=1e-7)


def test_map_heights_from_coordinates(raster_mapper_full):
    # Tested with gdal -> gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 556538.9876 429158.5652
    # Report:
    # Location: (469.512998180733P, 430.567498733493L)
    # Band 1:
    # Value: 1284.07670385579
    coo = np.array(
        [
            [556538.9876, 429158.5652, 1284.07670385579],
            [556538.9876, 429158.5652, 1284.07670385579],
        ]
    )
    result = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert result.ok is True
    assert np.allclose(result.coordinates, coo, atol=0.01, rtol=0)

    # Different CRS
    coo_utm = np.array(
        [
            [531550.98696, 5287808.89400, 1289.12332],
            [531547.64402920, 5287818.13604243, 1284.07670385579],
        ]
    )
    crs = CRS("EPSG:25833+5778")

    result = raster_mapper_full.georef_array.map_heights_from_coordinates(coo_utm, crs)
    assert result.ok is True
    assert np.allclose(result.coordinates, coo_utm, atol=0.01, rtol=0)

    # Edge case on the upper limit -> width of raster = 1011

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561950.0 429158.5652
    # Location: (1010.40023729484P,430.567498733493L)
    #   Value: 1024.56193042452
    #   coordinates in raster crs
    coo = np.array([561950.0, 429158.5652])
    coo_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 1024.56193042452, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561955.0 429158.5652
    #   Location: (1010.90003954914P,430.567498733493L)
    #   Value: 1024.73956631826
    coo = np.array([561955.0, 429158.5652])
    coo_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 1024.73956631826, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561956.0 429158.5652
    #   Location: (1011P,430.567498733493L)
    # Location is off this file! No further details to report.
    coo = np.array([561956.0, 429158.5652])
    coo_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert coo_map_heights.ok is False

    # Edge case on the lower limit ->
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551842.0 429158.5652
    #   Location: (0P,430.567498733493L)
    #    Value: 952.09822577356
    # coordinates in raster crs
    coo = np.array([551842.0, 429158.5652])
    coo_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 952.09822577356, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551845.0 429158.5652
    # Report:
    #   Location: (0.299881352577358P,430.567498733493L)
    #   Band 1:
    #     Value: 952.09822577356
    coo = np.array([551845.0, 429158.5652])
    coo_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 952.09822577356, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551847.5 429158.5652
    # Report:
    #   Location: (0.549782479727583P,430.567498733493L)
    #   Band 1:
    #     Value: 952.275286000538
    coo = np.array([551847.5, 429158.5652])
    coo_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 952.275286000538, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551845.0 433464.5652
    #  Location: (0.299881352577358P,0.143421380911605L)
    #  Value: 1102.26696777344
    coo = np.array([551845.0, 433464.5652])
    coo_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 1102.26696777344, atol=1e-6, rtol=0)


def test_map_heights_vertical_equal(raster_mapper_full):
    # Tested with gdal -> gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 556538.9876 429158.5652
    # Report:
    # Location: (469.512998180733P, 430.567498733493L)
    # Band 1:
    # Value: 1284.07670385579

    height_ref = 1284.07670385579

    # That test is for both crs are the same
    coo = np.array([[556538.9876, 429158.5652, 2000]])
    ray_vec = np.array([[0.0, 0.0, -1.0]])

    result_vertical_ray = raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )

    result_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)

    assert result_map_heights.ok is True
    assert result_vertical_ray.ok is True
    assert np.allclose(
        result_vertical_ray.coordinates,
        result_map_heights.coordinates,
        atol=0.001,
        rtol=0,
    )
    assert np.isclose(height_ref, result_map_heights.coordinates[0, 2], atol=0.001, rtol=0)

    # That test is for different crs systems
    coo = np.array([[531547.64402920, 5287818.13604243, 2000]])
    crs = CRS("EPSG:25833+5778")
    ray_vec = np.array([[0.0, 0.0, -1.0]])

    result_vertical_ray = raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=crs
    )

    result_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, crs)

    assert result_map_heights.ok is True
    assert result_vertical_ray.ok is True
    assert np.allclose(
        result_vertical_ray.coordinates,
        result_map_heights.coordinates,
        atol=0.001,
        rtol=0,
    )
    assert np.isclose(height_ref, result_vertical_ray.coordinates[0, 2], atol=0.01, rtol=0)

    # Edge case on the upper limit -> width of raster = 1011
    # These are valid position for the ray intersection
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561950.0 429158.5652
    # Location: (1010.40023729484P,430.567498733493L)
    #   Value: 1024.56193042452
    #   coordinates in raster crs
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551847.5 429158.5652
    #   Location: (0.549782479727583P,430.567498733493L)
    #    Value: 952.275286000538

    height_ref = np.array([1024.56193042, 952.275286000538])

    coo = np.array([[561950.0, 429158.5652, 8000], [551847.5, 429158.5652, 8000]])

    ray_vec = np.array([[0.0, 0.0, -1.0]])
    rays = np.zeros(coo.shape, dtype=float) + ray_vec

    result_map_heights = raster_mapper_full.georef_array.map_heights_from_coordinates(coo, raster_mapper_full.crs)
    result_vertical_ray = raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=rays, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )
    assert result_map_heights.ok is True
    assert result_vertical_ray.ok is True
    assert np.allclose(
        result_vertical_ray.coordinates,
        result_map_heights.coordinates,
        atol=0.001,
        rtol=0,
    )
    assert np.allclose(height_ref, result_vertical_ray.coordinates[:, 2], atol=0.01, rtol=0)

    # This are vertical rays little bit inside the allowed border
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561950.9980217605 429158.5652
    #   Location: (1010.49999999999P,430.567498733493L)
    #   Value: 1024.73956631824
    coo = np.array([[561950.9980217605, 429158.5652, 6000]])
    ray_vec = np.array([[0.0, 0.0, -1.0]])

    result_vertical_ray = raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )

    assert result_vertical_ray.ok is True
    assert np.allclose(1024.73956631824, result_vertical_ray.coordinates[:, 2], atol=1e-7, rtol=0)

    # This are vertical rays little bit outside the allowed border
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561950.9980217605 429158.5652
    #   Location: (1010.49999999999P,430.567498733493L)
    #   Value: 1024.73956631824
    coo = np.array([[561950.998021761, 429158.5652, 6000]])
    ray_vec = np.array([[0.0, 0.0, -1.0]])

    coo_map_heights = raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )
    assert coo_map_heights.ok is False

    # This are vertical rays within the raster but to close to the border where
    # the bilinear interpolation of a ray is not working anymore as the gdal bilinear interpolation where
    # For pixle coordinates lower 0.5 just 0.5 is used and similar for the max values
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561955.0 429158.5652
    #   Location: (1010.90003954914P,430.567498733493L)
    #   Value: 1024.73956631826
    coo = np.array([[561955.0, 429158.5652, 6000]])
    ray_vec = np.array([[0.0, 0.0, -1.0]])
    raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )
    assert coo_map_heights.ok is False

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551845.0 429158.5652
    #   Location: (0.299881352577358P,430.567498733493L)
    #   Value: 952.09822577356
    coo = np.array([[551845.0, 429158.5652, 6000]])
    raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )
    assert coo_map_heights.ok is False

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551845.0 433464.5652
    #  Location: (0.299881352577358P,0.143421380911605L)
    #  Value: 1102.26696777344
    coo = np.array([[551845.0, 433464.5652, 6000]])

    raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )
    assert coo_map_heights.ok is False
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551842.0 429158.5652
    #   Location: (0P,430.567498733493L)
    #    Value: 952.09822577356
    coo = np.array([[551842.0, 429158.5652, 6000]])
    raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )
    assert coo_map_heights.ok is False
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561956.0 429158.5652
    #   Location: (1011P,430.567498733493L)
    # Location is off this file! No further details to report.
    coo = np.array([[561956.0, 429158.5652, 6000]])
    raster_mapper_full.georef_array.map_coordinates_from_rays(
        ray_vectors_crs_s=ray_vec, ray_start_crs_s=coo, crs_s=raster_mapper_full.crs
    )
    assert coo_map_heights.ok is False



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

from pathlib import Path

import numpy as np
import pytest
from pyproj import CRS
from pyproj import network as network_pyproj

from weitsicht.exceptions import CRSnoZaxisError, MappingError
from weitsicht.mapping.raster import MappingRaster, MappingType

DATA_DIR = Path(__file__).parent.resolve() / "data"

network_pyproj.set_network_enabled(True)  # type: ignore


@pytest.fixture
def raster():
    # First test wrong datafiles:
    # Test wrong path
    with pytest.raises(FileNotFoundError):
        MappingRaster(DATA_DIR / "random_jpg_wrong.JPG")

    # Test non supported file by rasterio
    with pytest.raises(MappingError):
        MappingRaster(__file__)

    # TODO test wrong CRS. Not sure how to do as the Mapping Raster awaits a CRS class

    # Test CRS system without z Axis
    with pytest.raises(CRSnoZaxisError):
        MappingRaster(DATA_DIR / "dhm_at_lamb_10m_2018.tif")

    # Test error if proload full and window are sprecified
    with pytest.raises(ValueError):
        MappingRaster(
            DATA_DIR / "dhm_at_lamb_10m_2018.tif",
            crs=CRS("EPSG:31287+5778"),
            preload_full_raster=True,
            preload_window=(556530, 429150, 556550, 429200),
        )

    # Test error window sprecified is not the correct length
    with pytest.raises(ValueError):
        MappingRaster(
            DATA_DIR / "dhm_at_lamb_10m_2018.tif",
            crs=CRS("EPSG:31287+5778"),
            preload_window=(429150, 556550, 429200),  # type: ignore
        )

    # Test error crs and force_no_crs are used together
    with pytest.raises(ValueError):
        MappingRaster(
            DATA_DIR / "dhm_at_lamb_10m_2018.tif",
            crs=CRS("EPSG:31287+5778"),
            force_no_crs=True,
        )

    # Test wrong index band
    with pytest.raises(ValueError):
        MappingRaster(DATA_DIR / "dhm_at_lamb_10m_2018.tif", index_band=2, force_no_crs=True)

    # Test file has more than one band no band index was specified
    with pytest.raises(ValueError):
        MappingRaster(DATA_DIR / "random_jpg.JPG", force_no_crs=True)

    # Test wrong geo-transform of raster
    with pytest.raises(MappingError):
        MappingRaster(DATA_DIR / "random_jpg.JPG", index_band=1)

    mapping_raster = MappingRaster(DATA_DIR / "dhm_at_lamb_10m_2018.tif", crs=CRS("EPSG:31287+5778"))

    assert int(mapping_raster.type)

    return mapping_raster


def test_type(raster):
    assert raster.type == MappingType.Raster


def test_from_dict():

    with pytest.raises(KeyError):
        MappingRaster.from_dict(
            {
                "raster_filepath_wrong": (DATA_DIR / "dhm_at_lamb_10m_2018.tif").as_posix(),
                "crs": CRS("EPSG:31287+5778").to_wkt(),
            }
        )

    MappingRaster.from_dict(
        {
            "type": "Raster",
            "raster_filepath": (DATA_DIR / "dhm_at_lamb_10m_2018.tif").as_posix(),
            "crs": CRS("EPSG:31287+5778").to_wkt(),
        }
    )


def test_param_dict(raster):
    assert raster.param_dict == {
        "type": "Raster",
        "raster_filepath": (DATA_DIR / "dhm_at_lamb_10m_2018.tif").as_posix(),
        "crs": CRS("EPSG:31287+5778").to_wkt(),
        "band": 1,
    }


def test_resolution(raster):
    assert np.isclose(raster.resolution, 10.004021836097209)


def test_transform(raster):
    assert True


def test_width(raster):
    assert True


def test_height(raster):
    assert True


def test_load_window():

    # Tested with gdal -> gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 556538.9876 429158.5652
    # Report:
    # Location: (469.512998180733P, 430.567498733493L)
    # Band 1:
    # Value: 1284.07670385579

    # Test window limits are wrong
    with pytest.raises(MappingError):
        MappingRaster(
            DATA_DIR / "dhm_at_lamb_10m_2018.tif",
            crs=CRS("EPSG:31287+5778"),
            preload_window=(-556530, 429150, 556550, 429200),
        )

    mapping_raster = MappingRaster(
        DATA_DIR / "dhm_at_lamb_10m_2018.tif",
        crs=CRS("EPSG:31287+5778"),
        preload_window=(556530, 429150, 556550, 429200),
    )

    assert mapping_raster.georef_array is not None
    result = mapping_raster.georef_array.map_heights_from_coordinates(
        np.array([556538.9876, 429158.5652]), crs_s=mapping_raster.crs
    )
    assert result.ok is True
    assert np.allclose(
        result.coordinates,
        np.array([[556538.9876, 429158.5652, 1284.07670386]]),
        atol=1e-7,
        rtol=0,
    )


def test_get_window_georef_array(raster):
    assert True


def test_pixel_to_coordinate(raster):
    assert True


def test_pixel_valid(raster):
    assert True


def test_coordinate_on_raster(raster):
    assert True


def test_coordinate_to_pixel(raster):
    assert True


def test_get_coordinate_height(raster):
    assert True


def test_intersection_ray(raster):
    assert True


def test_map_coordinates_from_rays(raster):
    point_3d = np.array([[557057, 429465.0, 1520]])

    ray_vec = np.array([[0.3, -0.3, -0.80]])

    ray_crs = CRS("EPSG:31287+5778")
    x = raster.map_coordinates_from_rays(ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs)

    # No we also test the case we use another CRS
    point_3d = np.array([[557057, 429465.0, 1520 - 156.68]])

    ray_vec = np.array([[0.3, -0.3, -0.80]])

    ray_crs = CRS("EPSG:31287+8881")
    x1 = raster.map_coordinates_from_rays(ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=ray_crs)

    assert np.allclose(x.coordinates, x1.coordinates + np.array([0, 0, 156.68]), atol=1e-4, rtol=0)

    # test now rays not intersecting because of wrong direction
    point_3d = np.array([[557057, 429465.0, 1520]])

    ray_vec = np.array([[0.3, -0.3, 0.80]])

    x = raster.map_coordinates_from_rays(ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=raster.crs)
    assert x.ok is False

    # test now rays not intersecting because of wrong direction
    # but add second ray with intersection
    point_3d = np.array([[557057, 429465.0, 1520], [557057, 429465.0, 1520]])
    ray_vec = np.array([[0.3, -0.3, 0.80], [0.3, -0.3, -0.80]])

    x = raster.map_coordinates_from_rays(ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=raster.crs)
    assert x.ok is True
    np.testing.assert_array_equal(x.mask, np.array([False, True]))
    # parallel lines should be in this case every horizontal line
    point_3d = np.array([[557057, 429465.0, 1520]])
    ray_vec = np.array([[1.0, 0.3, 0.0]])

    x = raster.map_coordinates_from_rays(ray_vectors_crs_s=ray_vec, ray_start_crs_s=point_3d, crs_s=raster.crs)
    assert x.ok is False


def test_map_heights_from_coordinates(raster):
    # Tested with gdal -> gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 556538.9876 429158.5652
    # Report:
    # Location: (469.512998180733P, 430.567498733493L)
    # Band 1:
    # Value: 1284.07670385579
    coo = np.array(
        [
            [556538.9876, 429158.5652, 1284.07670385579],
            [556538.9876, 429158.5652, 1284.07670385579],
        ]
    )
    result = raster.map_heights_from_coordinates(coo, raster.crs)
    assert np.allclose(result.coordinates, coo, atol=0.01, rtol=0)

    # Different CRS
    coo_utm = np.array(
        [
            [531550.98696, 5287808.89400, 1289.12332],
            [531547.64402920, 5287818.13604243, 1284.07670385579],
        ]
    )
    crs = CRS("EPSG:25833+5778")

    result = raster.map_heights_from_coordinates(coo_utm, crs)
    assert result.ok is True
    assert np.allclose(result.coordinates, coo_utm, atol=0.01, rtol=0)

    # Edge case on the upper limit -> width of raster = 1011

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561950.0 429158.5652
    # Location: (1010.40023729484P,430.567498733493L)
    # Value: 1024.56193042452
    # coordinates in raster crs
    coo = np.array([561950.0, 429158.5652])
    coo_map_heights = raster.map_heights_from_coordinates(coo, raster.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 1024.56193042452, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561955.0 429158.5652
    #   Location: (1010.90003954914P,430.567498733493L)
    #   Value: 1024.73956631826
    coo = np.array([561955.0, 429158.5652])
    coo_map_heights = raster.map_heights_from_coordinates(coo, raster.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 1024.73956631826, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 561956.0 429158.5652
    # Location: (1011P,430.567498733493L)
    # Location is off this file! No further details to report.
    coo = np.array([561956.0, 429158.5652])

    coo_map_heights = raster.map_heights_from_coordinates(coo, raster.crs)
    assert coo_map_heights.ok is False

    # Edge case on the lower limit ->
    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551842.0 429158.5652
    #   Location: (0P,430.567498733493L)
    #    Value: 952.09822577356
    # coordinates in raster crs
    coo = np.array([551842.0, 429158.5652])
    coo_map_heights = raster.map_heights_from_coordinates(coo, raster.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 952.09822577356, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551845.0 429158.5652
    # Report:
    #   Location: (0.299881352577358P,430.567498733493L)
    #   Band 1:
    #     Value: 952.09822577356
    coo = np.array([551845.0, 429158.5652])
    coo_map_heights = raster.map_heights_from_coordinates(coo, raster.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 952.09822577356, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551847.5 429158.5652
    # Report:
    #   Location: (0.549782479727583P,430.567498733493L)
    #   Band 1:
    #     Value: 952.275286000538
    coo = np.array([551847.5, 429158.5652])
    coo_map_heights = raster.map_heights_from_coordinates(coo, raster.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 952.275286000538, atol=1e-6, rtol=0)

    # gdallocationinfo -r bilinear -geoloc dhm_at_lamb_10m_2018.tif 551845.0 433464.5652
    #  Location: (0.299881352577358P,0.143421380911605L)
    #  Value: 1102.26696777344
    coo = np.array([551845.0, 433464.5652])
    coo_map_heights = raster.map_heights_from_coordinates(coo, raster.crs)
    assert coo_map_heights.ok is True
    assert np.allclose(coo_map_heights.coordinates[0, 2], 1102.26696777344, atol=1e-6, rtol=0)


def test_map_heights_from_coordinates2(raster):
    coo = np.array([[555837, 430141], [555840, 430141], [555836, 430150], [561555.930, 428812.474]])

    result = raster.map_heights_from_coordinates(coo, raster.crs)

    assert result.ok is True
    assert np.allclose(result.coordinates[:, :2], coo, atol=0.01, rtol=0)

    vienna_heights_crs = CRS("EPSG:31287+8881")
    result_2 = raster.map_heights_from_coordinates(coo, crs_s=vienna_heights_crs)
    assert result.ok is True
    assert np.allclose(
        result.coordinates,
        result_2.coordinates + np.array([0, 0, 156.68]),
        atol=1e-6,
        rtol=0,
    )



from pathlib import Path

import numpy as np
import pytest

from weitsicht import Issue
from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective
from weitsicht.image.image_batch import ImageBatch
from weitsicht.image.perspective import ImagePerspective
from weitsicht.mapping.horizontal_plane import MappingHorizontalPlane
from weitsicht.mapping.map_trimesh import MappingTrimesh
from weitsicht.transform.rotation import Rotation

DATA_DIR = Path(__file__).parent.resolve() / "data"


@pytest.fixture
def images():
    image_dict = {}

    cam_20mm = CameraOpenCVPerspective(
        width=8256,
        height=5504,
        fx=4678.5675170822651,
        fy=4678.5675170822651,
        cx=8256 / 2.0 + 72.226251723716217 - 0.5,
        cy=5504 / 2.0 - 1.6581994715748369 - 0.5,
        k1=-0.043081934066152315,
        k2=0.013751556689262834,
        p1=-0.00075253724902191611,
        p2=0.0027883695798851496,
    )

    mapper_wall = MappingTrimesh(DATA_DIR / "decimated_160k_31256.ply")

    ext_ori_images = [
        [
            "INDIGO_2021-12-17_Z7II-B_0039",
            2643.0560,
            342715.0911648543551564,
            4.9515,
            -58.407,
            85.239,
            149.603,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0040",
            2643.2077,
            342717.0722784474492073,
            4.9169,
            -53.782,
            86.843,
            144.631,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0041",
            2643.1326,
            342719.1589016793295741,
            4.9015,
            -20.583,
            83.681,
            110.981,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0042",
            2642.9176,
            342720.8983379695564508,
            4.9114,
            -80.291,
            58.317,
            173.197,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0043",
            2645.3547,
            342718.4010614864528179,
            5.0342,
            -102.517,
            82.314,
            -166.5470,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0044",
            2642.7455,
            342718.9792082067579031,
            5.0487,
            102.283,
            87.869,
            79.249,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0045",
            2642.3706,
            342717.4982716664671898,
            5.0628,
            7.087,
            88.263,
            174.190,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0046",
            2642.4335,
            342715.4646163340657949,
            5.0715,
            -58.419,
            86.871,
            -121.629,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0047",
            2642.5897,
            342713.6570004606619477,
            5.0697,
            -125.144,
            88.884,
            -54.450,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0048",
            2642.7148,
            342713.2602515472099185,
            4.9537,
            81.508,
            57.961,
            5.612,
        ],
    ]

    for eor in ext_ori_images:
        position = np.array([eor[1], eor[2], eor[3]])
        orientation = Rotation.from_opk_degree(eor[4], eor[5], eor[6])
        image = ImagePerspective(
            width=8256,
            height=5504,
            position=position,
            orientation=orientation,
            crs=None,
            camera=cam_20mm,
            mapper=mapper_wall,
        )
        image_dict[eor[0]] = image

    images = ImageBatch(image_dict)

    return images


def test_add_images(images):
    image_dict = {}

    cam_20mm = CameraOpenCVPerspective(
        width=8256,
        height=5504,
        fx=4678.5675170822651,
        fy=4678.5675170822651,
        cx=8256 / 2.0 + 72.226251723716217 - 0.5,
        cy=5504 / 2.0 - 1.6581994715748369 - 0.5,
        k1=-0.043081934066152315,
        k2=0.013751556689262834,
        p1=-0.00075253724902191611,
        p2=0.0027883695798851496,
    )

    ext_ori_images = [
        [
            "INDIGO_2021-12-17_Z7II-B_0039",
            2643.0560,
            342715.0911648543551564,
            4.9515,
            -58.407,
            85.239,
            149.603,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0040",
            2643.2077,
            342717.0722784474492073,
            4.9169,
            -53.782,
            86.843,
            144.631,
        ],
    ]

    for eor in ext_ori_images:
        position = np.array([eor[1], eor[2], eor[3]])
        orientation = Rotation.from_opk_degree(eor[4], eor[5], eor[6])
        image = ImagePerspective(
            width=8256,
            height=5504,
            position=position,
            orientation=orientation,
            crs=None,
            camera=cam_20mm,
        )
        image_dict[eor[0]] = image

    with pytest.raises(KeyError):
        images.add_images(image_dict)

    # Now add other names
    image_dict = {}
    ext_ori_images = [
        [
            "INDIGO_2021-12-17_Z7II-B_0039_2",
            2643.0560,
            342715.0911648543551564,
            4.9515,
            -58.407,
            85.239,
            149.603,
        ],
        [
            "INDIGO_2021-12-17_Z7II-B_0040_2",
            2643.2077,
            342717.0722784474492073,
            4.9169,
            -53.782,
            86.843,
            144.631,
        ],
    ]

    for eor in ext_ori_images:
        position = np.array([eor[1], eor[2], eor[3]])
        orientation = Rotation.from_opk_degree(eor[4], eor[5], eor[6])
        image = ImagePerspective(
            width=8256,
            height=5504,
            position=position,
            orientation=orientation,
            crs=None,
            camera=cam_20mm,
        )
        image_dict[eor[0]] = image

    images.add_images(image_dict)
    assert len(images) == 12


def test___setter(images):

    with pytest.raises(ValueError):
        images["new_image"] = "string is not an image"

    images["new_image"] = images.index(0)
    assert len(images) == 11


def test___getter(images):

    res = images[["INDIGO_2021-12-17_Z7II-B_0039", "INDIGO_2021-12-17_Z7II-B_0040"]]
    assert len(res) == 2

    with pytest.raises(KeyError):
        images["wrong_image_key"]


def test_index(images):
    with pytest.raises(IndexError):
        images.index(100)

    res = images.index([1, 3])
    assert len(res) == 2


def test_empty_image_batch_exceptions():
    images = ImageBatch(images={})

    with pytest.raises(ValueError):
        images.map_center_point()

    with pytest.raises(ValueError):
        images.map_footprint()

    with pytest.raises(ValueError):
        images.project(np.array([[0, 0, 0]]))


def test_map_center_point(images):
    result = images.map_center_point()

    assert len(result) == 10

    res_image = result["INDIGO_2021-12-17_Z7II-B_0039"]
    assert res_image.ok is True
    assert np.isclose(res_image.gsd, 0.001, atol=0.0001, rtol=0.0)
    assert np.allclose(
        res_image.coordinates,
        np.array([2.63808856e03, 3.42714739e05, 4.73475821e00]),
        atol=0.001,
        rtol=0.0,
    )

    # We will add 3 images
    # the new images will point away from the mesh
    # but the first one has no camera model
    # the second one has no mapper specified
    # the third should be returning false because of wrong direction
    image_dict = {}
    ext_ori_images = [
        [
            "fake_1",
            2643.0560,
            342715.0911648543551564,
            4.9515,
            -58.407,
            -85.239,
            149.603,
            None,
            None,
        ],
        [
            "fake_2",
            2643.2077,
            342717.0722784474492073,
            4.9169,
            -53.782,
            -86.843,
            144.631,
            images.index(0).camera,
            None,
        ],
        [
            "fake_3",
            2643.2077,
            342717.0722784474492073,
            4.9169,
            -53.782,
            -86.843,
            144.631,
            images.index(0).camera,
            images.index(0).mapper,
        ],
    ]

    for eor in ext_ori_images:
        position = np.array([eor[1], eor[2], eor[3]])
        orientation = Rotation.from_opk_degree(eor[4], eor[5], eor[6])
        image = ImagePerspective(
            width=8256,
            height=5504,
            position=position,
            orientation=orientation,
            mapper=eor[8],
            camera=eor[7],
        )
        image_dict[eor[0]] = image

    images.add_images(image_dict)

    result = images.map_center_point()

    assert len(result) == 13
    assert result["fake_1"].ok is False
    assert Issue.IMAGE_BATCH_ERROR in result["fake_1"].issues
    assert Issue.IMAGE_BATCH_ERROR in result["fake_1"].issues
    assert result["fake_2"].ok is False
    assert result["fake_3"].ok is False
    assert Issue.NO_INTERSECTION in result["fake_3"].issues

    images_2 = ImageBatch({"a": images.index(0)})
    # Plane with altitude over all other
    # None of the center points can be mapped
    mapper = MappingHorizontalPlane(8000, crs=None)
    result = images_2.map_center_point(mapper=mapper)
    for _, res in result.items():
        assert res.ok is False


def test_map_footprint(images):
    # We change the position to be closer to the wall
    # just because the mesh is very tigh cut on the edges. And the image was part of the mesh creation
    # So for some areas where the image may have no overlap there would be no mesh
    images["INDIGO_2021-12-17_Z7II-B_0044"]._position = np.array([2638.7455, 342718.9792082067579031, 5.0487])

    result = images.map_footprint()

    assert len(result) == 10

    res_image = result["INDIGO_2021-12-17_Z7II-B_0044"]
    assert res_image.ok is True
    assert np.isclose(res_image.gsd, 0.00025, atol=0.0001, rtol=0.0)
    assert np.allclose(
        res_image.coordinates[0],
        np.array([2.63788572e03, 3.42718522e05, 4.25969046e00]),
        atol=0.001,
        rtol=0.0,
    )

    # Plane with altitude over all other
    mapper = MappingHorizontalPlane(8000, crs=None)

    map_result = images.map_footprint(mapper=mapper)
    for _, res in map_result.items():
        assert res.ok is False


def test_project(images):
    projection_dict = images.project(np.array([2637.874263, 342718.336006, 6.225628]))
    assert projection_dict is not None
    assert len(projection_dict) == 10

    projection_dict = images.project(np.array([2637.874263, 342718.336006, 6.225628]), only_valid=True)
    assert projection_dict is not None
    assert len(projection_dict) == 8

    # No valid projections
    projection_dict = images.project(np.array([2637.874263, 342718.336006, 80000]), only_valid=True)
    assert projection_dict is None

    # Fake that all images have no valid georeference
    for _, v in images.images.items():
        v._position = None
    # No valid projections
    projection_dict = images.project(np.array([2637.874263, 342718.336006, 80000]), only_valid=True)
    assert projection_dict is None



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""3D coplanarity checks."""

__all__ = ["is_coplanar"]

# Python program to check if 4 points
# in a 3-D plane are Coplanar


# Function to find equation of plane.
def is_coplanar(p1, p2, p3, p_test):
    """Return whether four 3D points are coplanar.

    Points are interpreted as array-like objects of length 3: ``(x, y, z)``.

    :param p1: First point.
    :type p1: array-like
    :param p2: Second point.
    :type p2: array-like
    :param p3: Third point.
    :type p3: array-like
    :param p_test: Fourth point to test.
    :type p_test: array-like
    :return: ``True`` if the four points are coplanar, otherwise ``False``.
    :rtype: bool
    """
    a1 = p2[0] - p1[0]
    b1 = p2[1] - p1[1]
    c1 = p2[2] - p1[2]
    a2 = p3[0] - p1[0]
    b2 = p3[1] - p1[1]
    c2 = p3[2] - p1[2]
    a = b1 * c2 - b2 * c1
    b = a2 * c1 - a1 * c2
    c = a1 * b2 - b1 * a2
    d = -a * p1[0] - b * p1[1] - c * p1[2]

    # equation of plane is: a*x + b*y + c*z = 0 #

    # checking if the 4th point satisfies
    # the above equation
    if a * p_test[0] + b * p_test[1] + c * p_test[2] + d == 0:
        return True
    else:
        return False



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Helpers for converting coordinates to GeoJSON geometries."""

from shapely import geometry

from weitsicht.utils import ArrayNx3, Vector3D


def get_geojson(coordinates: Vector3D | ArrayNx3, geom_type: str) -> dict:
    """Create a GeoJSON geometry dictionary for coordinates of a given type.

    Supported geometry types are ``Point``, ``LineString`` and ``Polygon``.

    :param coordinates: Coordinates used to construct the geometry.
    :type coordinates: Vector3D | ArrayNx3
    :param geom_type: Geometry type string.
    :type geom_type: str
    :return: GeoJSON geometry mapping as returned by :func:`shapely.geometry.mapping`.
    :rtype: dict
    :raises ValueError: If ``geom_type`` is unsupported.
    :raises Exception: If Shapely cannot construct the requested geometry from ``coordinates``.
    """

    if geom_type == "Point":
        geom = geometry.Point(coordinates)
    elif geom_type == "LineString":
        geom = geometry.LineString(coordinates)
    elif geom_type == "Polygon":
        geom = geometry.Polygon(coordinates)
    else:
        raise ValueError(r"Geometry is not supported. Only 'Point', 'LineString', 'Polygon'")

    geojson = geometry.mapping(geom)

    return geojson



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------
"""Mapping backends and factory helpers.

The mapping package contains mapper implementations (raster, mesh, plane, …) and utilities
for constructing mappers from configuration dictionaries.
"""

import logging

from weitsicht.mapping.base_class import MappingBase, MappingType
from weitsicht.mapping.georef_array import MappingGeorefArray
from weitsicht.mapping.horizontal_plane import MappingHorizontalPlane
from weitsicht.mapping.map_trimesh import MappingTrimesh
from weitsicht.mapping.mapping_dict_selector import get_mapper_from_dict
from weitsicht.mapping.raster import MappingRaster

__all__ = [
    "MappingBase",
    "MappingType",
    "MappingGeorefArray",
    "MappingHorizontalPlane",
    "MappingTrimesh",
    "MappingRaster",
    "get_mapper_from_dict",
]

logger = logging.getLogger(__name__)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------


"""Orthophoto image model backed by rasterio geotransforms."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import cast

import numpy as np
import pyproj.exceptions
import rasterio
from affine import Affine
from pyproj import CRS
from rasterio import errors
from shapely import geometry

from weitsicht.exceptions import (
    CRSInputError,
    MapperMissingError,
    NotGeoreferencedError,
)
from weitsicht.image.base_class import ImageBase, ImageType
from weitsicht.mapping.base_class import MappingBase
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.utils import (
    ArrayNx2,
    ArrayNx3,
    Issue,
    MappingResult,
    MappingResultSuccess,
    MaskN_,
    ProjectionResult,
    ProjectionResultSuccess,
    ResultFailure,
    Vector3D,
    to_array_nx2,
)

__all__ = ["ImageOrtho"]

logger = logging.getLogger(__name__)


class ImageOrtho(ImageBase):
    """Orthophoto image model based on an affine geotransform.

    This image type uses an affine transform to convert between image pixel coordinates and world
    coordinates on a plane (x/y). Heights can optionally be provided via a mapper.
    """

    def __init__(
        self,
        width: float | int,
        height: float | int,
        geo_transform: rasterio.Affine | None = None,
        resolution: float = 0.0,
        crs: CRS | None = None,
        mapper: MappingBase | None = None,
    ):
        """Initialize an orthophoto image model.

        :param width: Image width in pixels.
        :type width: float | int
        :param height: Image height in pixels.
        :type height: float | int
        :param geo_transform: Affine geotransform, defaults to ``None``.
        :type geo_transform: rasterio.Affine | None
        :param resolution: Pixel resolution in world units, defaults to ``0.0``. If ``0.0`` and a geotransform is
            provided, the mean scaling of the transform is used.
        :type resolution: float
        :param crs: World CRS of the orthophoto, defaults to ``None``.
        :type crs: CRS | None
        :param mapper: Mapping instance used for height queries, defaults to ``None``.
        :type mapper: MappingBase | None
        """
        super().__init__(mapper=mapper, crs=crs, width=width, height=height)

        # geo transform
        self.geo_transform = geo_transform

        self._res = resolution

        if self._res == 0.0:
            if self.geo_transform is not None:
                # TODO all that resolution could be optimized, problem can arise if non-cartesian CRS is given
                # for the ortho imagery
                self._res = float(np.mean(self.geo_transform._scaling))

    @classmethod
    def from_file(cls, path: Path | str, crs: CRS | None = None, mapper: MappingBase | None = None) -> ImageOrtho:
        """Create an orthophoto image model from a raster file.

        :param path: Path to the raster file.
        :type path: Path | str
        :param crs: CRS which overrides the file CRS, defaults to ``None``.
        :type crs: CRS | None
        :param mapper: Mapping instance, defaults to ``None``.
        :type mapper: MappingBase | None
        :return: Image model instance.
        :rtype: ImageOrtho
        :raises Exception: If rasterio cannot read the file or required metadata is unavailable.
        """

        if not Path(path).exists():
            raise FileNotFoundError("Specified file does not exist")
        path = Path(path)

        dataset = rasterio.open(path)
        width = dataset.shape[0]
        height = dataset.shape[1]

        # rasterio returns identity if file has no geo-reference
        # We want that gt is None instead of having an identity matrix
        if dataset.transform.is_identity:
            gt = None
        else:
            gt = dataset.transform

        _crs: CRS | None = crs
        if crs is None:
            # get crs code from orthophoto and promote to 3d.
            # Actually that would not be needed for the orthophoto
            # but to be sure all pyproj transformations are working
            if dataset.crs is not None:
                _crs = CRS(dataset.crs).to_3d()

        # TODO this should be optimized
        if dataset.crs is not None:
            try:
                linear_unit_factor = dataset.crs.linear_units_factor
                res = float(np.mean(dataset.res) * linear_unit_factor[1])
            except errors.CRSError:
                linear_unit_factor = 1.0
                res = float(np.mean(dataset.res) * linear_unit_factor)

        else:
            res = float(np.mean(dataset.res))

        # initialize the image class
        image = cls(width, height, geo_transform=gt, resolution=res, crs=_crs, mapper=mapper)

        return image

    @property
    def resolution(self) -> float:
        """Return the pixel resolution in world units.

        :return: Resolution value.
        :rtype: float
        """
        return self._res

    @property
    def type(self):
        """Return the image model type.

        :return: Image type.
        :rtype: ImageType
        """
        return ImageType.Orthophoto

    @classmethod
    def from_dict(cls, param_dict: dict, mapper: MappingBase | None = None) -> ImageOrtho:
        """Create an :class:`ImageOrtho` from a parameter dictionary.

        Required keys are:
        - ``width``
        - ``height``

        Optional keys are:
        - ``geo_transform`` (Affine as dict)
        - ``resolution``
        - ``crs`` (WKT)

        :param param_dict: Dictionary with image parameters.
        :type param_dict: dict
        :param mapper: Mapping instance, defaults to ``None``.
        :type mapper: MappingBase | None
        :return: Image model instance.
        :rtype: ImageOrtho
        :raises KeyError: If required keys are missing.
        :raises ValueError: If values are invalid.
        :raises TypeError: If values have incompatible types.
        :raises CRSInputError: If the CRS WKT string is invalid or unsupported.
        """

        width = param_dict["width"]
        height = param_dict["height"]

        # get geo transform
        gt = None
        if param_dict.get("geo_transform", None) is not None:
            gt = rasterio.Affine(**param_dict["geo_transform"])

        resolution = param_dict.get("resolution", 0.0)

        crs = None
        if param_dict.get("crs", False):
            try:
                crs = CRS(param_dict["crs"])
            except pyproj.exceptions.CRSError as err:
                raise CRSInputError("No valid CRS WKT string supported") from err

        image = cls(
            width,
            height,
            geo_transform=gt,
            resolution=resolution,
            crs=crs,
            mapper=mapper,
        )

        return image

    @property
    def param_dict(self) -> dict:
        """Return image parameters as a dictionary.

        The returned dictionary is compatible with :meth:`from_dict`.

        :return: Image parameters.
        :rtype: dict
        """
        param_dict = {
            "type": self.type.fullname,
            "width": self._width,
            "height": self._height,
            "geo_transform": self.geo_transform._asdict() if self.geo_transform is not None else None,
            "resolution": self._res,
        }
        if self.crs is not None:
            param_dict["crs"] = self.crs.to_wkt()

        return param_dict

    @property
    def center(self) -> tuple[float, float]:
        """Return the image center in pixel coordinates.

        :return: Center point (x, y) in pixel coordinates.
        :rtype: tuple[float, float]
        """
        return self.width / 2.0, self.height / 2.0

    @property
    def is_geo_referenced(self) -> bool:
        """Return whether the image is geo-referenced.

        :return: ``True`` if a geotransform is set, otherwise ``False``.
        :rtype: bool
        """
        if self.geo_transform is not None:
            return True
        return False

    def image_points_inside(self, point_image_coordinates: ArrayNx2) -> MaskN_:
        """Return whether image pixel coordinates are inside the raster bounds.

        :param point_image_coordinates: Image pixel coordinates (image CRS).
        :type point_image_coordinates: ArrayNx2
        :return: Boolean mask of shape ``(N,)``.
        :rtype: ``MaskN_``
        :raises ValueError: If the input cannot be converted to an ``(N, 2)`` array.
        :raises TypeError: If the input has incompatible types.
        """

        # Format points to be numpy array Nx2
        point_image_coordinates = to_array_nx2(point_image_coordinates)

        min_border = np.logical_and(point_image_coordinates[:, 0] >= 0, point_image_coordinates[:, 1] >= 0)
        max_border = np.logical_and(
            point_image_coordinates[:, 0] < (self.width - 1.0),
            point_image_coordinates[:, 1] < (self.height - 1.0),
        )
        valid_index = np.logical_and(min_border, max_border)

        return valid_index

    def position_to_crs(self, crs_t: CRS, transformer: CoordinateTransformer | None = None) -> Vector3D | None:
        """Return the orthophoto center position in the target CRS.

        If a mapper is available, the Z value is obtained by mapping the center point.

        :param crs_t: Target CRS.
        :type crs_t: CRS
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Center position in ``crs_t`` or ``None`` if unavailable.
        :rtype: Vector3D | None
        :raises CRSInputError: If CRS/transformer input is invalid.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        if self.geo_transform is None:
            return None

        point_center_2d = self.geo_transform * [self.width / 2.0, self.height / 2.0]

        point_center_3d = np.array([[point_center_2d[0], point_center_2d[1], 0]])

        if self.mapper is not None:
            crs_arg = None if transformer is not None else self._crs
            mapping_result = self.mapper.map_heights_from_coordinates(point_center_2d, crs_arg, transformer=transformer)

            if mapping_result.ok is True:
                point_center_3d = mapping_result.coordinates

        # Transformer
        if self._crs is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=self._crs, crs_t=crs_t)
        )

        if coo_trafo is not None:
            coordinates_crs_t = coo_trafo.transform(point_center_3d)
        else:
            coordinates_crs_t = np.array([[point_center_3d[0][0], point_center_3d[0][1], point_center_3d[0][2]]])

        return coordinates_crs_t[0, :]

    def project(
        self, coordinates: ArrayNx3 | ArrayNx2, crs_s: CRS | None = None, transformer: CoordinateTransformer | None = None
    ) -> ProjectionResult:
        """Project 3D coordinates into orthophoto pixel coordinates.

        Once the coordinates from the coordinates crs (crs_s) are transformed to the mapper crs,
        then onlyx/y coordinates are used for the affine transform; z is ignored.

        The reason of the tranformatoin from crs_s to mapper crs is why also here for orthophoto,
        3D coordiantes must be used as input.

        :param coordinates: 3D coordinates to project.
        :type coordinates: ArrayNx3
        :param crs_s: CRS of the input coordinates, defaults to ``None``.
        :type crs_s: CRS | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Projection result.
        :rtype: ProjectionResult
        :raises ValueError: If CRS/transformer inputs are inconsistent.
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises CRSInputError: If CRS/transformer input is invalid.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        # _coordinates = to_array_nx3(coordinates)

        # Transformer
        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        if coo_trafo is not None:
            point_3d_crs = coo_trafo.transform(coordinates)
        else:
            point_3d_crs = coordinates * 1.0

        assert self.geo_transform is not None

        gt = cast(Affine, ~self.geo_transform)  # invert
        x, y = cast(tuple[np.ndarray, np.ndarray], gt * (point_3d_crs[:, 0], point_3d_crs[:, 1]))
        # x, y = ~self.geo_transform * (point_3d_crs[:, 0], point_3d_crs[:, 1])

        raster_coordinates = np.vstack((x, y)).T
        valid_mask = self.image_points_inside(raster_coordinates)

        issue = set()
        if not np.all(valid_mask):
            issue = {Issue.INVALID_PROJECTIIONS}
            if not np.any(valid_mask):
                return ResultFailure(
                    ok=False,
                    error="None of the coordinates are on the raster",
                    issues={Issue.INVALID_PROJECTIIONS},
                )

        return ProjectionResultSuccess(ok=True, pixels=raster_coordinates, mask=valid_mask, issues=issue)

    def map_center_point(
        self, mapper: MappingBase | None = None, transformer: CoordinateTransformer | None = None
    ) -> MappingResult:
        """Map the orthophoto center pixel to 3D coordinates via a mapper.

        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`~weitsicht.image.base_class.ImageBase.mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises MapperMissingError: If no mapper is available.
        :raises CRSInputError: If the mapper rejects CRS/transformer input.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        :raises WeitsichtError: Base class for all weitsicht exceptions. Catch this to handle any weitsicht error;
            catch specific subclasses first if you need to distinguish causes.
        """

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        if self.mapper is None and mapper is None:
            raise MapperMissingError("No mapper provided for map_points")

        mapper_to_use = self.mapper
        if mapper is not None:
            mapper_to_use = mapper

        assert mapper_to_use is not None
        assert self.geo_transform is not None

        point_center_2d = self.geo_transform * [self.width / 2.0, self.height / 2.0]
        mapping_result = mapper_to_use.map_heights_from_coordinates(
            point_center_2d, None if transformer is not None else self._crs, transformer=transformer
        )

        if mapping_result.ok is False:
            return ResultFailure(ok=False, error=mapping_result.error, issues=mapping_result.issues)

        return MappingResultSuccess(
            ok=True,
            coordinates=mapping_result.coordinates,
            mask=mapping_result.mask,
            gsd=self.resolution,
            issues=mapping_result.issues,
        )

    def map_footprint(
        self,
        points_per_edge: int = 0,
        mapper: MappingBase | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map the orthophoto footprint polygon to 3D coordinates via a mapper.

        :param points_per_edge: Number of points inserted between corners, defaults to ``0``.
            Values ``< 1`` map only the corner points.
        :type points_per_edge: int
        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`~weitsicht.image.base_class.ImageBase.mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises MapperMissingError: If no mapper is available.
        :raises CRSInputError: If the mapper rejects CRS/transformer input.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        :raises WeitsichtError: Base class for all weitsicht exceptions. Catch this to handle any weitsicht error;
            catch specific subclasses first if you need to distinguish causes.
        """

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        if self.mapper is None and mapper is None:
            raise MapperMissingError("No mapper provided for map_points")

        mapper_to_use = self.mapper
        if mapper is not None:
            mapper_to_use = mapper

        assert self.geo_transform is not None
        assert mapper_to_use is not None

        if points_per_edge < 1:
            px = np.array([0.0, 0.0, float(self.width), float(self.width)], dtype=np.float64)
            py = np.array([0.0, float(self.height), float(self.height), 0.0], dtype=np.float64)
        else:
            footprint_points_2d = [[x, 0] for x in np.linspace(0, self.width, 2 + points_per_edge)]
            footprint_points_2d += [[self.width, x] for x in np.linspace(0, self.height, 2 + points_per_edge)][1:]
            footprint_points_2d += [[x, self.height] for x in np.linspace(self.width, 0, 2 + points_per_edge)][1:]
            footprint_points_2d += [[0, x] for x in np.linspace(self.height, 0, 2 + points_per_edge)][1:-1]
            _fp = np.array(footprint_points_2d)
            px = _fp[:, 0]
            py = _fp[:, 1]

        pt_x, pt_y = cast(tuple[np.ndarray, np.ndarray], self.geo_transform * (px, py))
        footprint_points_2d = np.vstack((pt_x, pt_y)).T
        mapping_result = mapper_to_use.map_heights_from_coordinates(
            footprint_points_2d, None if transformer is not None else self._crs, transformer=transformer
        )

        if mapping_result.ok is False:
            return ResultFailure(ok=False, error=mapping_result.error, issues=mapping_result.issues)

        if not np.all(mapping_result.mask):
            return ResultFailure(
                ok=False,
                error="Some footprint points failed",
                issues=mapping_result.issues,
            )

        # TODO Implementation of resolution and area could be a problem for non cartesian georef of ortho
        footprint_geom = geometry.Polygon(mapping_result.coordinates)
        area = float(np.round(footprint_geom.area))
        return MappingResultSuccess(
            ok=True,
            coordinates=mapping_result.coordinates,
            mask=mapping_result.mask,
            gsd=self.resolution,
            area=area,
            issues=mapping_result.issues,
        )

    def map_points(
        self,
        points_image: ArrayNx2 | ArrayNx3 | list[list[float]] | list[list[int]] | list[float] | list[int],
        mapper: MappingBase | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map orthophoto pixel coordinates to 3D coordinates via a mapper.

        :param points_image: Pixel coordinates (image CRS).
        :type points_image: ArrayNx2 | ArrayNx3 | list[list[float]] | list[list[int]] | list[float] | list[int]
        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`~weitsicht.image.base_class.ImageBase.mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises ValueError: If ``points_image`` cannot be parsed as an array of 2D points.
        :raises TypeError: If ``points_image`` has incompatible types.
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises MapperMissingError: If no mapper is available.
        :raises CRSInputError: If the mapper rejects CRS/transformer input.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        :raises WeitsichtError: Base class for all weitsicht exceptions. Catch this to handle any weitsicht error;
            catch specific subclasses first if you need to distinguish causes.
        """

        # TODO what should we do if specified points_image are outside the orthophoto
        # but the mapping still works? Should be masked them as invalid?

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        if self.mapper is None and mapper is None:
            raise MapperMissingError("No mapper provided for map_points")

        mapper_to_use = self.mapper
        if mapper is not None:
            mapper_to_use = mapper

        assert mapper_to_use is not None
        assert self.geo_transform is not None

        _points_image = to_array_nx2(points_image)

        pt_x, pt_y = cast(tuple[np.ndarray, np.ndarray], self.geo_transform * (_points_image[:, 0], _points_image[:, 1]))

        mapping_result = mapper_to_use.map_heights_from_coordinates(
            np.vstack((pt_x, pt_y)).T, None if transformer is not None else self._crs, transformer=transformer
        )

        if mapping_result.ok is False:
            return ResultFailure(ok=False, error=mapping_result.error, issues=mapping_result.issues)

        # TODO Implementation of resolution and area could be a problem for non cartesian georef of ortho
        return MappingResultSuccess(
            ok=True,
            coordinates=mapping_result.coordinates,
            mask=mapping_result.mask,
            gsd=self.resolution,
            issues=mapping_result.issues,
        )



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Deprecated coordinate transformation utilities.

This module contains a legacy transformer implementation kept for information.
For new code, prefer :class:`weitsicht.transform.coordinates_transformer.CoordinateTransformer`.
"""

from __future__ import annotations

import logging
from functools import lru_cache

import numpy as np
from pyproj import CRS, Transformer
from pyproj.exceptions import CRSError, ProjError
from shapely import errors as shapely_errors
from shapely import geometry

from weitsicht import cfg
from weitsicht.exceptions import CoordinateTransformationError
from weitsicht.utils import ArrayNx2, ArrayNx3, Vector2D, Vector3D

logger = logging.getLogger(__name__)


# The doc states that cached Versions can be used to speed up performance significantly
TransformerFromCRS = lru_cache(Transformer.from_crs)
TransformerFromPipeline = lru_cache(Transformer.from_pipeline)


class CoordinatesTransformer:  # pragma: no cover
    """Legacy coordinate transformation wrapper around :mod:`pyproj`.

    The coordinates of this class are read-only once initialized.
    """

    def __init__(self, crs: CRS, points: ArrayNx3 | Vector3D | Vector2D | ArrayNx2):
        """Create a transformer with CRS and coordinates.

        :param crs: CRS of the input coordinates.
        :type crs: CRS
        :param points: Coordinates to transform.
        :type points: ArrayNx3 | Vector3D | Vector2D | ArrayNx2
        """

        self._coordinates = points.copy()

        # The class uses internal only array with ndim 2. So point np.array([0,0,1]) will be np.array([[0,0,1]])
        if self._coordinates.ndim == 1:
            self._coordinates = np.array([self._coordinates])

        # self._coordinates.setflags(write=False)
        self._crs: CRS = crs

        self.transformer = None

    @classmethod
    def from_crs(
        cls, crs_s: CRS, crs_t: CRS, points: ArrayNx3 | Vector3D | Vector2D | ArrayNx3
    ) -> CoordinatesTransformer:
        """Create and immediately transform coordinates from ``crs_s`` to ``crs_t``.

        :param crs_s: Source CRS.
        :type crs_s: CRS
        :param crs_t: Target CRS.
        :type crs_t: CRS
        :param points: Coordinates to transform.
        :type points: ArrayNx3 | Vector3D | Vector2D | ArrayNx3
        :return: Transformer instance in target CRS.
        :rtype: CoordinatesTransformer
        :raises CoordinateTransformationError: If establishing or applying the transformation fails.
        """

        _obj = cls(crs=crs_s, points=points)

        return _obj.to_crs(crs_t)

    @property
    def is_point(self) -> bool:
        if self._coordinates.shape[0] == 1:
            return True
        return False

    @property
    def is_3d(self):
        if self._coordinates.shape[1] == 3:
            return True
        return False

    def geojson(self, geom_type: str) -> dict:
        """Convert the stored coordinates into a GeoJSON mapping.

        :param geom_type: Geometry type (``Point``, ``LineString``, ``Polygon``).
        :type geom_type: str
        :return: GeoJSON mapping dict.
        :rtype: dict
        :raises CoordinateTransformationError: If geometry type is unsupported or coordinates are invalid.
        """

        try:
            if geom_type == "Point":
                geojson = geometry.Point(self.coordinates)
            elif geom_type == "LineString":
                geojson = geometry.LineString(self._coordinates)
            elif geom_type == "Polygon":
                geojson = geometry.Polygon(self._coordinates)
            else:
                raise CoordinateTransformationError(r"Geometry is not supported. Only 'Point', 'LineString', 'Polygon'")

        except (shapely_errors.GEOSException, ValueError) as err:
            # Will also raise if multi coordinates and point type is used
            raise CoordinateTransformationError("GeoJSON can not be established from given coordinates") from err

        geojson = geometry.mapping(geojson)

        return geojson

    @property
    def coordinates(self) -> ArrayNx3 | ArrayNx2:
        return self._coordinates

    @coordinates.setter
    def coordinates(self, new_coordinate: ArrayNx3 | ArrayNx2):

        points = new_coordinate.copy()

        # The class uses internal only array with ndim 2. So point np.array([0,0,1]) will be np.array([[0,0,1]])
        if points.ndim == 1:
            new_coordinate = np.array([new_coordinate])
        self._coordinates = new_coordinate

    def to_crs(self, crs_target: CRS) -> CoordinatesTransformer:
        """Transform the stored coordinates to ``crs_target``.

        :param crs_target: Target CRS.
        :type crs_target: CRS
        :return: New transformer instance holding coordinates in ``crs_target``.
        :rtype: CoordinatesTransformer
        :raises CoordinateTransformationError: If establishing or applying the transformation fails.
        """

        # If the same coordinate systems are used we will spare the trafo
        if self._crs.equals(crs_target):
            return CoordinatesTransformer(crs_target, self._coordinates)

        try:
            self.transformer = TransformerFromCRS(
                self._crs,
                crs_target,
                always_xy=True,
                allow_ballpark=cfg._ballpark_transformation,
                only_best=cfg._only_best_transformation,
            )
            # if cfg._only_best_transformation and not tr_group.best_available:
            #    raise CoordinateTransformationError("Best Transformation is not available "
            #                                        "but only best transformations are allowed")
            # self.transformer = tr_group.transformers[0]

            # self.transformer = Transformer.from_crs(self._crs, crs_target, always_xy=True,
            #                                        only_best=cfg._only_best_transformation,
            #                                        allow_ballpark=cfg._ballpark_transformation)
        except ProjError as e:
            raise CoordinateTransformationError("Transformation could not be established.") from e

        try:
            if self.is_3d:
                fx, fy, fz = self.transformer.transform(
                    self._coordinates[:, 0],
                    self._coordinates[:, 1],
                    self._coordinates[:, 2],
                    errcheck=True,
                )
                p_crs = np.vstack((fx, fy, fz)).T
            else:
                fx, fy = self.transformer.transform(self._coordinates[:, 0], self._coordinates[:, 1], errcheck=True)
                p_crs = np.vstack((fx, fy)).T

            new_trafo = CoordinatesTransformer(crs_target, p_crs)
            new_trafo.transformer = self.transformer
            return new_trafo

        except (CRSError, ProjError) as e:
            raise CoordinateTransformationError("Error while transforming coordinates") from e



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------


import numpy as np
import pytest
from pyproj import CRS

import weitsicht
from weitsicht.transform.utm_converter import get_zone, point_convert_utm_wgs84_egm2008


def test_get_zone():
    utm_zone = get_zone(longitude=16, latitude=48)
    assert utm_zone == 32633


def test_get_zone_fails():
    with pytest.raises(ValueError):
        get_zone(longitude=200, latitude=48)


def test_wgs84_utm_point_converter():
    weitsicht.allow_ballpark_transformations()
    weitsicht.allow_non_best_transformations()
    x_utm, y_utm, z_geoid, utm_crs = point_convert_utm_wgs84_egm2008(CRS(4979), x=16, y=18, z=100)

    # We test only for x and y, Z is configured to use geoid data which depending on your interpreters
    # pyproj settings could have the geoid file
    assert np.linalg.norm(np.array([x_utm, y_utm]) - np.array([(605866.998, 1990471.052)])) < 0.001
    assert utm_crs.to_wkt() == CRS("EPSG:32633+3855").to_wkt()

    x_utm, y_utm, z_geoid, utm_crs = point_convert_utm_wgs84_egm2008(
        CRS("EPSG:32633"), x=605866.998, y=1990471.052, z=100
    )



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""UTM conversion helpers."""

from __future__ import annotations

import logging

import numpy as np
from pyproj import CRS
from pyproj.crs.crs import CompoundCRS

from weitsicht.transform.coordinates_transformer import CoordinateTransformer

__all__ = ["get_zone", "point_convert_utm_wgs84_egm2008"]

logger = logging.getLogger(__name__)


def get_zone(longitude: float, latitude: float) -> int:
    """Estimate the WGS84 utm zone from latitude and longitude.

    :param longitude: Longitude in degrees. Valid range: ``-180`` to ``180``.
    :type longitude: float
    :param latitude: Latitude in degrees. Valid range: ``-90`` to ``90``.
    :type latitude: float
    :return: UTM EPSG code.
    :rtype: int
    :raises ValueError: If limits are exceeded.
    """

    if abs(longitude) > 180 or abs(latitude) > 90:
        raise ValueError("Limits exceeded: -180<Longitude<180 and -90<Latitude<90")

    epsg_code_utm = int(32700 - np.round((45 + latitude) / 90, 0) * 100 + np.round((183 + longitude) / 6, 0))
    return epsg_code_utm


def point_convert_utm_wgs84_egm2008(
    crs_s: CRS, x: float, y: float, z: float
) -> tuple[float, float, float, CRS | CompoundCRS]:
    """Transform a single point into WGS84-UTM (EGM2008) coordinates.

    The point is first transformed to WGS84 3D (EPSG:4979), then assigned to a UTM zone and
    transformed to the corresponding compound CRS (UTM + EGM2008 geoid height).

    :param crs_s: CRS of the input point.
    :type crs_s: CRS
    :param x: X coordinate in ``crs_s`` units.
    :type x: float
    :param y: Y coordinate in ``crs_s`` units.
    :type y: float
    :param z: Z coordinate in ``crs_s`` units.
    :type z: float
    :return: Tuple ``(x_utm, y_utm, z_geoid, utm_crs)``.
    :rtype: tuple[float, float, float, CRS | CompoundCRS]
    :raises ValueError: If the transformed WGS84 point is outside UTM latitude/longitude limits.
    :raises CoordinateTransformationError: If a coordinate transformation cannot be established or applied.
    """

    crs_4979 = CRS(4979)

    coo_trafo = CoordinateTransformer.from_crs(crs_s, crs_4979)

    if coo_trafo is not None:
        coo_wgs84 = coo_trafo.transform(np.array([x, y, z]))
    else:
        coo_wgs84 = np.array([[x, y, z]])

    epsg_code_utm = get_zone(coo_wgs84[0][0], coo_wgs84[0][1])

    utm_crs = CRS("EPSG:" + str(epsg_code_utm) + "+3855")

    # Here the trafo can actually never be None
    transformer_wgs84 = CoordinateTransformer.from_crs(crs_4979, utm_crs)

    assert transformer_wgs84 is not None
    coo_utm = transformer_wgs84.transform(*coo_wgs84)

    x_utm, y_utm, z_geoid = coo_utm[0, :]
    return x_utm, y_utm, z_geoid, utm_crs



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------


"""Tag system registry and helpers.

Each tag system lives in its own module with a concrete resolver class.
"""

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any


@dataclass
class MetaTagIORBase:
    image_shape: tuple[int, int]
    focal_plane_resolution_unit: int | None
    focal_plane_x_resolution: float | None
    focal_plane_y_resolution: float | None
    make: str
    model: str
    focal_length: float | None
    focal_length_35mm: float | None


@dataclass
class MetaTagGPS:
    gps_latitude: float | None
    gps_latitude_ref: str | None
    gps_longitude: float | None
    gps_longitude_ref: str | None
    gps_altitude: float | None


@dataclass
class MetaTagCRS:
    horiz_cs: str | None
    vert_cs: str | None


@dataclass
class MetaTagZalternatives:
    rel_altitude: float | None


@dataclass
class MetaTagOrientation:
    xmp_camera_pitch: float | None
    xmp_camera_roll: float | None
    xmp_camera_yaw: float | None

    maker_notes_camera_pitch: float | None
    maker_notes_camera_roll: float | None
    maker_notes_camera_yaw: float | None

    xmp_pitch: float | None
    xmp_roll: float | None
    xmp_yaw: float | None

    maker_notes_pitch: float | None
    maker_notes_roll: float | None
    maker_notes_yaw: float | None

    xmp_gimbal_roll_deg: float | None
    xmp_gimbal_yaw_deg: float | None
    xmp_gimbal_pitch_deg: float | None

    maker_notes_gimbal_roll_deg: float | None
    maker_notes_gimbal_yaw_deg: float | None
    maker_notes_gimbal_pitch_deg: float | None


@dataclass
class MetaTagIORExtended:
    image_shape: tuple[int, int]
    calibrated_focal_length: float | None
    calibrated_optical_center_x: float | None
    calibrated_optical_center_y: float | None


@dataclass
class MetaTagAll:
    """Container holding all resolved tag groups."""

    ior_base: MetaTagIORBase
    ior_extended: MetaTagIORExtended
    gps: MetaTagGPS
    orientation: MetaTagOrientation
    crs: MetaTagCRS
    z_alternatives: MetaTagZalternatives


class MetaTagsBase:
    """Base class how different meta-data parser should look like"""

    def __init__(self, meta_data: Any):
        self.meta_data = meta_data

    @abstractmethod
    def image_shape(self) -> tuple[int, int]:

        return (0, 0)

    @abstractmethod
    def get_ior_base(self) -> MetaTagIORBase:

        tags = MetaTagIORBase(
            image_shape=self.image_shape(),
            focal_plane_resolution_unit=0,
            focal_plane_x_resolution=0,
            focal_plane_y_resolution=0,
            make="",
            model="",
            focal_length=0.0,
            focal_length_35mm=0.0,
        )
        return tags

    @abstractmethod
    def get_standard_gps(self) -> MetaTagGPS:

        tags = MetaTagGPS(
            gps_latitude=None,
            gps_latitude_ref=None,
            gps_longitude=None,
            gps_longitude_ref=None,
            gps_altitude=None,
        )

        return tags

    @abstractmethod
    def get_crs(self) -> MetaTagCRS:
        tags = MetaTagCRS(horiz_cs=None, vert_cs=None)

        return tags

    @abstractmethod
    def get_z_alternatives(self) -> MetaTagZalternatives:
        tags = MetaTagZalternatives(rel_altitude=None)

        return tags

    @abstractmethod
    def get_orientation_values(self) -> MetaTagOrientation:

        tags = MetaTagOrientation(
            xmp_camera_pitch=None,
            xmp_camera_roll=None,
            xmp_camera_yaw=None,
            maker_notes_camera_pitch=None,
            maker_notes_camera_roll=None,
            maker_notes_camera_yaw=None,
            xmp_pitch=None,
            xmp_roll=None,
            xmp_yaw=None,
            maker_notes_pitch=None,
            maker_notes_roll=None,
            maker_notes_yaw=None,
            xmp_gimbal_roll_deg=None,
            xmp_gimbal_yaw_deg=None,
            xmp_gimbal_pitch_deg=None,
            maker_notes_gimbal_roll_deg=None,
            maker_notes_gimbal_yaw_deg=None,
            maker_notes_gimbal_pitch_deg=None,
        )
        return tags

    def get_ior_extended(self) -> MetaTagIORExtended:

        tags = MetaTagIORExtended(
            image_shape=self.image_shape(),
            calibrated_focal_length=None,
            calibrated_optical_center_x=None,
            calibrated_optical_center_y=None,
        )

        return tags

    def get_all(self) -> MetaTagAll:

        tags = MetaTagAll(
            ior_base=self.get_ior_base(),
            ior_extended=self.get_ior_extended(),
            gps=self.get_standard_gps(),
            orientation=self.get_orientation_values(),
            z_alternatives=self.get_z_alternatives(),
            crs=self.get_crs(),
        )

        return tags



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Bilinear interpolation utilities."""

__all__ = ["bilinear_interpolation"]


def bilinear_interpolation(x: float | int, y: float | int, points: tuple | list) -> float:
    """Bilinear interpolation on a rectangle defined by four corner samples.

    The ``points`` must be a sequence of four ``(x, y, value)`` tuples forming a rectangle.

    :param x: X coordinate where the value should be interpolated.
    :type x: float | int
    :param y: Y coordinate where the value should be interpolated.
    :type y: float | int
    :param points: Four corner points ``(x, y, value)``.
    :type points: tuple | list
    :return: Interpolated value.
    :rtype: float
    :raises ValueError: If the points do not form a rectangle or if ``(x, y)`` lies outside the rectangle.
    :raises TypeError: If values have incompatible types.



        >>> bilinear_interpolation(12, 5.5,
        ...                        [(10, 4, 100),
        ...                         (20, 4, 200),
        ...                         (10, 6, 150),
        ...                         (20, 6, 300)])
        165.0

    """
    # See formula at:  http://en.wikipedia.org/wiki/Bilinear_interpolation

    points = sorted(points)  # order points by x, then by y
    (x1, y1, q11), (_x1, y2, q12), (x2, _y1, q21), (_x2, _y2, q22) = points

    if x1 != _x1 or x2 != _x2 or y1 != _y1 or y2 != _y2:
        raise ValueError("points do not form a rectangle")
    if not x1 <= x <= x2 or not y1 <= y <= y2:
        raise ValueError("(x, y) not within the rectangle")

    return (
        q11 * (x2 - x) * (y2 - y) + q21 * (x - x1) * (y2 - y) + q12 * (x2 - x) * (y - y1) + q22 * (x - x1) * (y - y1)
    ) / ((x2 - x1) * (y2 - y1))



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

import inspect
import logging

from weitsicht.camera.base_perspective import CameraBasePerspective
from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective
from weitsicht.metadata.tag_systems.tag_base import MetaTagIORBase, MetaTagIORExtended

logger = logging.getLogger(__name__)


class AlternativeCalibrationTags:
    """Class which holds function to estimate camera calibration of different types using metadata
    All methods implemented will be called at once. So if you want to extend different tagging systems
    from different software packages or vendors just add a new method

    This is only for the camera calibration aka IOR. Exterior orientation aka EOR will be treated in another class
    This should be unique. All methods will be tried, the first one which is not returning None is used.
    """

    def __init__(self, tags: MetaTagIORBase, tags_ior_extended: MetaTagIORExtended):
        self.tags_ior = tags
        self.tags_ior_extended = tags_ior_extended
        self.width = int(self.tags_ior.image_shape[0])
        self.height = int(self.tags_ior.image_shape[1])

    def dji_tags(self) -> CameraOpenCVPerspective | None:

        focal_pixel = None
        c_y = self.height / 2.0
        c_x = self.width / 2.0
        focal_val = self.tags_ior_extended.calibrated_focal_length
        if focal_val is not None:
            focal_pixel = float(focal_val)

        # For some DjI images calibrated tags have been found
        c_x_val = self.tags_ior_extended.calibrated_optical_center_x
        c_y_val = self.tags_ior_extended.calibrated_optical_center_y
        if c_x_val is not None and c_y_val is not None:
            c_x = float(c_x_val)
            c_y = float(c_y_val)

        # Focal Pixel is not found than do not use that even if Calibrated Center would be found
        if focal_pixel is None:
            return None

        camera = CameraOpenCVPerspective(
            width=self.width,
            height=self.height,
            fx=focal_pixel,
            fy=focal_pixel,
            cx=c_x,
            cy=c_y,
        )
        return camera

    def pix4d_tags(self):

        # TODO implement as described in https://support.pix4d.com/hc/en-us/articles/205732309
        return None

    @staticmethod
    def run_all_methods(tags_ior: MetaTagIORBase, tags_ior_extended: MetaTagIORExtended) -> CameraBasePerspective | None:

        v = AlternativeCalibrationTags(tags_ior, tags_ior_extended)
        attrs = (getattr(v, name) for name in dir(v) if not name.startswith("_"))
        methods = filter(inspect.ismethod, attrs)

        cam_result = None
        for method in methods:
            # print(method)
            try:
                cam_result = method()
                if cam_result is not None:
                    break
            except Exception:
                logger.warning("One of the Calibration Methods throws an exception")
                pass

        return cam_result



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Ray intersection with a bilinear patch."""

import numpy as np
from numpy.polynomial import Polynomial

from weitsicht.utils import ArrayNx3, Vector3D

__all__ = ["multilinear_poly_intersection"]


def multilinear_poly_intersection(points: ArrayNx3, p: Vector3D, r: Vector3D) -> Vector3D | None:
    """Calculate the intersection point of a ray with a bilinear polynom.
    The ray is defined by a point p and a direction r. The bilinear patch is defined by 4 points
    It is assumed that the points form a rectangle as this is used to test the validity of intersection point
    to be within the rectangle.

    :param points: The 4 corner points of the bilinear patch.
    :type points: ArrayNx3
    :param p: A point on the ray. The point location is important as for
                2 intersection solutions the closer solution to p will be used
    :type p: Vector3D
    :param r: The direction vector
    :type r: Vector3D
    :return: The first intersection point or None if no intersection is found within the limits of the points
    :rtype: Vector3D | None
    """

    # z = a00 + a01 * x + a10 * y + a11 * x * y
    # ray = p + r * t
    # z_ray = p_z + r_z * t
    # x_ray = p_x + r_x * t
    # y_ray = p_y + r_y * t

    # substitute z with z_ray and x,y of the bilinear polynom with x_ray and y_ray
    # will give us the quadratic equation for t

    # First we will solve the linear equation to get a00, a01, a10, a11
    lin_matrix = np.array([[1, 1, 1, 1], points[:, 0], points[:, 1], points[:, 0] * points[:, 1]]).T
    coeff = np.linalg.solve(lin_matrix, points[:, 2])
    a00, a01, a10, a11 = coeff

    # Second we will create the polynom of order 2
    poly = Polynomial(
        [
            -p[2] + a00 + a01 * p[0] + a10 * p[1] + a11 * p[0] * p[1],
            -r[2] + a01 * r[0] + a10 * r[1] + a11 * p[0] * r[1] + a11 * p[1] * r[0],
            a11 * r[0] * r[1],
        ]
    )

    # Finding the roots gives the value of t for the intersection points.
    roots = poly.roots()

    roots = roots[np.isreal(roots)]
    if len(roots) == 0:
        return None

    # Just to be sure. Maybe there are numerical instabilities
    roots = np.real(roots)
    p_solutions = p + np.outer(r, roots).T

    # points[:,:2].min(axis=0)<= p_solutions[:,:2]
    x_within = np.logical_and(
        (points[:, 0].min() - 1e-10) <= p_solutions[:, 0],
        p_solutions[:, 0] <= (points[:, 0].max() + 1e-10),
    )

    y_within = np.logical_and(
        (points[:, 1].min() - 1e-10) <= p_solutions[:, 1],
        p_solutions[:, 1] <= (points[:, 1].max() + 1e-10),
    )

    valid_solution_index = np.flatnonzero(np.logical_and(x_within, y_within))

    if valid_solution_index.size == 0:
        return None

    elif valid_solution_index.size == 1:
        return p_solutions[valid_solution_index, :][0]

    # More than one solution
    # Will use the closer on
    else:
        close_index = np.argmin(np.linalg.norm(p - p_solutions, axis=1))
        return p_solutions[close_index, :]



import json
from pathlib import Path

import numpy as np
from pyproj import CRS

import weitsicht
from weitsicht import ImageFromMetaBuilder, PyExifToolTags, eor_from_meta, image_from_meta

DATA_DIR = Path(__file__).parent.resolve() / "data"
DJI_P1_TAGS_JSON = DATA_DIR / "image_exiftool_tags.json"


def _load_meta() -> dict:
    with DJI_P1_TAGS_JSON.open(encoding="utf-8") as handle:
        return json.load(handle)


def test_eor_from_meta_zenmuse_p1():
    weitsicht.allow_ballpark_transformations()
    weitsicht.allow_non_best_transformations()

    tags = PyExifToolTags(_load_meta())
    result = eor_from_meta(tags.get_all())

    assert result.ok is True

    assert result.crs is not None
    assert result.crs.to_wkt() == CRS("EPSG:32636+3855").to_wkt()

    assert result.position.shape == (3,)
    assert np.all(np.isfinite(result.position))
    assert 0.0 < float(result.position[0]) < 1_000_000.0  # easting
    assert 0.0 < float(result.position[1]) < 10_000_000.0  # northing

    rot = result.orientation.matrix
    assert rot.shape == (3, 3)
    assert np.allclose(rot.T @ rot, np.eye(3), atol=1e-6)
    assert abs(float(np.linalg.det(rot)) - 1.0) < 1e-6


def test_image_from_meta_builder():
    weitsicht.allow_ballpark_transformations()
    weitsicht.allow_non_best_transformations()

    tags = PyExifToolTags(_load_meta())

    builder = ImageFromMetaBuilder(tags)
    ior_1 = builder.ior()
    ior_2 = builder.ior()
    assert ior_1 is ior_2
    assert ior_1.ok is True
    assert (ior_1.width, ior_1.height) == (8192, 5460)

    eor_1 = builder.eor()
    eor_2 = builder.eor()
    assert eor_1 is eor_2
    assert eor_1.ok is True

    img_builder = builder.image()
    assert img_builder.ok is True
    assert img_builder.image.is_geo_referenced is True
    assert (img_builder.image.width, img_builder.image.height) == (8192, 5460)
    assert img_builder.image.crs is not None

    img_func = image_from_meta(tags)
    assert img_func.ok is True
    assert img_func.image.is_geo_referenced is True
    assert (img_func.image.width, img_func.image.height) == (8192, 5460)
    assert img_func.image.crs is not None
    assert img_func.image.crs.to_wkt() == img_builder.image.crs.to_wkt()



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

import numpy
import pytest

from weitsicht.camera.base_perspective import CameraType
from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective


@pytest.fixture
def camera_class():
    return CameraOpenCVPerspective(
        width=5472,
        height=3648,
        fx=3753.0,
        fy=3754.0,
        cx=2721.2,
        cy=1813.7,
        k1=-0.082,
        k2=0.053,
        k3=0.0089,
        p1=-0.0015,
        p2=-0.00086,
    )


def test_camera_inti():

    with pytest.raises(ValueError):
        CameraOpenCVPerspective(width=0, height=3648, fx=3753.0, fy=3754.0)
    with pytest.raises(ValueError):
        CameraOpenCVPerspective(width=5472, height=0, fx=3753.0, fy=3754.0)
    with pytest.raises(ValueError):
        CameraOpenCVPerspective(width=5472, height=3648, fx=0.0, fy=3754.0)
    with pytest.raises(ValueError):
        CameraOpenCVPerspective(width=5472, height=3648, fx=3753.0, fy=0)

    camera = CameraOpenCVPerspective(width=5472, height=3648, fx=3753.0, fy=3753.0)
    assert camera.cx == 5472 / 2.0 - 0.5
    assert camera.cy == 3648 / 2.0 - 0.5


def test_from_dict(camera_class):
    cam = CameraOpenCVPerspective.from_dict(
        {
            "type": "OpenCV",
            "calib_width": 5472,
            "calib_height": 3648,
            "fx": 3753.0,
            "fy": 3754.0,
            "cx": 2721.2,
            "cy": 1813.7,
            "k1": -0.082,
            "k2": 0.053,
            "k3": 0.0089,
            "k4": 0.0,
            "p1": -0.0015,
            "p2": -0.00086,
        }
    )
    assert cam.param_dict == camera_class.param_dict

    with pytest.raises(KeyError):
        cam = CameraOpenCVPerspective.from_dict(
            {
                "type": "OpenCV",
                "calib_width": 5472,
                "calib_height": 3648,
                "fx1": 3753.0,
                "fy": 3754.0,
                "cx": 2721.2,
                "cy": 1813.7,
                "k1": -0.0082,
                "k2": 0.00053,
                "k3": 0.0089,
                "k4": 0.0,
                "p1": -0.0015,
                "p2": -0.00086,
            }
        )

    with pytest.raises(ValueError):
        cam = CameraOpenCVPerspective.from_dict(
            {
                "type": "OpenCV",
                "calib_width": 0,
                "calib_height": 3648,
                "fx": 3753.0,
                "fy": 3754.0,
                "cx": 2721.2,
                "cy": 1813.7,
                "k1": -0.0082,
                "k2": 0.00053,
                "k3": 0.0089,
                "k4": 0.0,
                "p1": -0.0015,
                "p2": -0.00086,
            }
        )


def test_type(camera_class):
    assert camera_class.type == CameraType.OpenCV


def test_param_dict(camera_class):
    assert camera_class.param_dict == {
        "type": "OpenCV",
        "calib_width": 5472,
        "calib_height": 3648,
        "fx": 3753.0,
        "fy": 3754.0,
        "cx": 2721.2,
        "cy": 1813.7,
        "k1": -0.082,
        "k2": 0.053,
        "k3": 0.0089,
        "k4": 0.0,
        "p1": -0.0015,
        "p2": -0.00086,
    }


def test_focal_length_for_gsd_in_pixel(camera_class):
    assert camera_class.focal_length_for_gsd_in_pixel == 3753.5


def test__principal_point(camera_class):
    assert camera_class._principal_point.tolist() == [camera_class.cx, camera_class.cy]


def test__camara_crs_to_image_crs(camera_class):
    result = camera_class._camara_crs_to_image_crs(numpy.array([0, 0, 1]))
    assert result[0].tolist() == [camera_class.cx, camera_class.cy]


def test__image_crs_to_vector_in_camera_crs(camera_class):
    result = camera_class._image_crs_to_vector_in_camera_crs(numpy.array([camera_class.cx, camera_class.cy]))
    assert result[0].tolist() == [0, 0, -1]


def test__to_distorted_points(camera_class):
    # The principal point will not be distorted
    result = camera_class._to_distorted_points(numpy.array([[camera_class.cx, camera_class.cy]]))

    assert result[0].tolist() == [camera_class.cx, camera_class.cy]


def test__to_undistorted_points(camera_class):
    result = camera_class._to_undistorted_points(numpy.array([[camera_class.cx, camera_class.cy]]))
    assert result[0].tolist() == [camera_class.cx, camera_class.cy]
    test_points = numpy.array([[10, 10]])
    result = camera_class._to_undistorted_points(test_points)
    result_2 = camera_class._to_distorted_points(result)
    assert numpy.linalg.norm(test_points - result_2) < 0.1



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Top-level package exports for ``weitsicht``.

This module re-exports the most commonly used classes, functions, and exceptions so that
user code can typically import from ``weitsicht`` directly.
"""

import logging
from importlib import metadata as importlibmetadata
from pathlib import Path

try:  # Python 3.11+
    import tomllib  # stdlib
except ModuleNotFoundError:  # Python 3.10
    import tomli as tomllib


from weitsicht import cfg
from weitsicht.camera import CameraType, get_camera_from_dict
from weitsicht.camera.base_perspective import CameraBasePerspective
from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective
from weitsicht.exceptions import (
    CoordinateTransformationError,
    CRSInputError,
    CRSnoZaxisError,
    MapperMissingError,
    MappingBackendError,
    MappingError,
    NotGeoreferencedError,
    WeitsichtError,
)
from weitsicht.geometry.coplanar_collinear import is_coplanar
from weitsicht.geometry.interpolation_bilinear import bilinear_interpolation
from weitsicht.geometry.intersection_bilinear import multilinear_poly_intersection
from weitsicht.geometry.intersection_plane import intersection_plane, intersection_plane_mat_operation
from weitsicht.geometry.line_grid_intersection import line_grid_intersection_points, raster_index_p1_p2, vector_projection
from weitsicht.image import get_image_from_dict
from weitsicht.image.base_class import ImageBase, ImageType
from weitsicht.image.image_batch import ImageBatch
from weitsicht.image.ortho import ImageOrtho
from weitsicht.image.perspective import ImagePerspective
from weitsicht.mapping import get_mapper_from_dict
from weitsicht.mapping.base_class import MappingBase, MappingType
from weitsicht.mapping.georef_array import MappingGeorefArray
from weitsicht.mapping.horizontal_plane import MappingHorizontalPlane
from weitsicht.mapping.map_trimesh import MappingTrimesh
from weitsicht.mapping.raster import MappingRaster
from weitsicht.metadata.camera_estimator_metadata import estimate_camera, ior_from_meta
from weitsicht.metadata.eor_from_meta import eor_from_meta
from weitsicht.metadata.image_from_meta import ImageFromMetaBuilder, image_from_meta
from weitsicht.metadata.tag_systems import PyExifToolTags
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.transform.rotation import Rotation
from weitsicht.type_guards import (
    is_camera_type,
    is_opencv_camera,
    is_ortho_image,
    is_perspective_image,
)
from weitsicht.utils import (
    Array3x3,
    ArrayN_,
    ArrayNx2,
    ArrayNx3,
    ArrayNxN,
    Issue,
    MappingResult,
    MappingResultSuccess,
    ProjectionResult,
    ProjectionResultSuccess,
    ResultFailure,
    Vector2D,
    Vector3D,
)

__all__ = [
    "CameraType",
    "get_camera_from_dict",
    "CameraOpenCVPerspective",
    "CameraBasePerspective",
    "ImageType",
    "get_image_from_dict",
    "ImagePerspective",
    "ImageBase",
    "ImageOrtho",
    "ImageBatch",
    "MappingType",
    "get_mapper_from_dict",
    "MappingBase",
    "MappingRaster",
    "MappingTrimesh",
    "MappingHorizontalPlane",
    "MappingGeorefArray",
    "Rotation",
    "CoordinateTransformer",
    "PyExifToolTags",
    "ior_from_meta",
    "estimate_camera",
    "image_from_meta",
    "ImageFromMetaBuilder",
    "eor_from_meta",
    "is_opencv_camera",
    "is_camera_type",
    "is_perspective_image",
    "is_ortho_image",
    "ArrayNx3",
    "ArrayNx2",
    "Vector2D",
    "Vector3D",
    "Array3x3",
    "ArrayN_",
    "ArrayNxN",
    "Issue",
    "is_coplanar",
    "bilinear_interpolation",
    "multilinear_poly_intersection",
    "intersection_plane",
    "intersection_plane_mat_operation",
    "vector_projection",
    "raster_index_p1_p2",
    "line_grid_intersection_points",
    "ProjectionResult",
    "MappingResult",
    "MappingResultSuccess",
    "ResultFailure",
    "ProjectionResultSuccess",
    "cfg",
    "allow_ballpark_transformations",
    "allow_non_best_transformations",
    "NotGeoreferencedError",
    "MapperMissingError",
    "MappingError",
    "MappingBackendError",
    "WeitsichtError",
    "CoordinateTransformationError",
    "CRSInputError",
    "CRSnoZaxisError",
]


def _read_pyproject_version() -> str:
    """Read the package version from ``pyproject.toml``.

    This is used as a fallback if the distribution metadata is not available (e.g. running
    from a source checkout without an installed wheel).

    :return: Version string.
    :rtype: str
    """
    pyproject = Path(__file__).resolve().parents[2] / "pyproject.toml"
    if not pyproject.is_file():
        return "0.0.0-dev"
    data = tomllib.loads(pyproject.read_text())
    if "project" in data and "version" in data["project"]:
        return data["project"]["version"]
    if "tool" in data and "poetry" in data["tool"] and "version" in data["tool"]["poetry"]:
        return data["tool"]["poetry"]["version"]
    return "0.0.0-dev"


try:
    __version__ = importlibmetadata.version("weitsicht")
except importlibmetadata.PackageNotFoundError:
    __version__ = _read_pyproject_version()
__author__ = "Martin Wieser"

logger = logging.getLogger(__name__)


def allow_ballpark_transformations(allow: bool = True) -> None:
    """Allow or disallow "ballpark" coordinate transformations.

    This toggles ``pyproj`` transformations that may use less-accurate but available
    fallbacks. Defaults to ``True``.

    :param allow: Whether to allow ballpark transformations, defaults to ``True``.
    :type allow: bool
    """
    cfg._ballpark_transformation = allow


def allow_non_best_transformations(allow: bool = True) -> None:
    """Allow or disallow non-best coordinate transformations.

    If set to ``False`` (default behavior in weitsicht), only "best available" transformations
    are used. Defaults to ``True``.

    :param allow: Whether to allow non-best transformations, defaults to ``True``.
    :type allow: bool
    """
    cfg._only_best_transformation = not allow



import numpy as np
import pytest

from weitsicht.utils import to_array_nx2, to_array_nx3


def test_to_array_nx2():

    with pytest.raises(ValueError):
        _ = to_array_nx2([100])

    with pytest.raises(ValueError):
        _ = to_array_nx2([[100], [100], [100]])

    np.testing.assert_allclose(to_array_nx2([100, 100, 100, 100]), np.array([[100, 100]]))


def test_to_array_nx3():

    with pytest.raises(ValueError):
        _ = to_array_nx3([100])

    with pytest.raises(ValueError):
        _ = to_array_nx3([[100], [100], [100]])

    np.testing.assert_allclose(to_array_nx2([100, 100, 100, 100]), np.array([[100, 100]]))

    with pytest.raises(ValueError):
        _ = to_array_nx3([100, 100])

    with pytest.raises(ValueError):
        _ = to_array_nx3([[100, 100], [100, 100], [100, 100]])

    arr = to_array_nx3([[100, 100], [100, 100], [100, 100]], fill_z=0)
    np.testing.assert_allclose(arr, np.array([[100, 100, 0], [100, 100, 0], [100, 100, 0]]))



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Mapping backend based on a :mod:`trimesh` mesh surface."""

from __future__ import annotations

import logging
from collections.abc import Mapping
from pathlib import Path
from typing import Any

import numpy as np
import pyproj.exceptions
import trimesh
from pyproj import CRS

from weitsicht.exceptions import CRSInputError, CRSnoZaxisError, MappingError
from weitsicht.geometry.ray_from_points import rays_from_points_batch
from weitsicht.mapping.base_class import MappingBase, MappingType
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.utils import (
    ArrayNx2,
    ArrayNx3,
    Issue,
    MappingResult,
    MappingResultSuccess,
    ResultFailure,
    Vector3D,
    to_array_nx3,
)

__all__ = ["MappingTrimesh"]

logger = logging.getLogger(__name__)
# attach to logger so trimesh messages will be printed to console
# trimesh.util.attach_to_log()


class MappingTrimesh(MappingBase):
    """Mapping backend that intersects rays with a mesh surface."""

    def __init__(
        self,
        mesh_path: Path | str,
        crs: CRS | None = None,
        coordinate_shift: Vector3D | None = None,
    ):
        """Create a mesh-based mapper.

        :param mesh_path: Path to the mesh file.
        :type mesh_path: Path | str
        :param crs: CRS of the mesh, defaults to ``None``.
        :type crs: CRS | None
        :param coordinate_shift: Optional translation applied to the mesh coordinates, defaults to ``None``.
        :type coordinate_shift: Vector3D | None
        :raises FileNotFoundError: If ``mesh_path`` does not exist.
        :raises CRSnoZaxisError: If a specified CRS does not define a Z axis.
        :raises MappingError: If the mesh is empty or the file format is unsupported.
        """
        super().__init__()

        self.path = Path(mesh_path)

        self.translation = coordinate_shift

        if not self.path.exists():
            raise FileNotFoundError(f"File specified not Found: {self.path.as_posix()}")

        try:
            # load_mesh needs Pillow to be installed, I was not able to shut that off
            # Thats the reason we are casting and checking that self._mnesh is a trimesh.Trimesh object
            # trimesh.load() returns GEOMETRY -> supercalls. but geometry has no ray function ->problem for linting
            self._mesh = trimesh.load(self.path)  # pyright: ignore[reportUnknownMemberType]
            assert isinstance(self._mesh, trimesh.Trimesh)
        except KeyError as e:
            raise MappingError(f"File format {self.path.suffix} not working for trimesh") from e
        except NotImplementedError as e:
            raise MappingError(f"File format {self.path.suffix} not working for trimesh") from e
        except ValueError as e:
            raise MappingError("File seems not working") from e

        if self._mesh.is_empty:
            raise MappingError("Trimesh object is empty")

        self._crs = crs

        if self._crs is not None:
            if len(self._crs.axis_info) < 3:
                logger.error("CRS has no Z axis defined")
                raise CRSnoZaxisError("CRS has no Z axis defined")

    @property
    def type(self):
        """Return the mapper type."""
        return MappingType.Trimesh

    @classmethod
    def from_dict(cls, mapper_dict: Mapping[str, Any]) -> MappingTrimesh:
        """Create a :class:`MappingTrimesh` instance from a configuration dictionary.

        Required keys:
        - ``type``: must be ``Trimesh``
        - ``mesh_filepath``: file path to the mesh

        Optional keys:
        - ``crs``: CRS WKT string, defaults to ``None``
        - ``coordinate_shift``: translation vector, defaults to ``None``

        :param mapper_dict: Mapper configuration dictionary (typically created via ``mapper.param_dict``).
        :type mapper_dict: Mapping[str, Any]
        :return: Instantiated mapper.
        :rtype: MappingTrimesh
        :raises KeyError: If a required dictionary key is missing.
        :raises ValueError: If the mapper type is unsupported.
        :raises CRSInputError: If the CRS WKT string is invalid.
        """

        try:
            type_text = mapper_dict["type"]
        except KeyError as err:
            raise KeyError("'type' key is missing") from err

        if type_text != MappingType.Trimesh.fullname:
            raise ValueError(f'"type" of mapper dict is not {MappingType.Trimesh.fullname}')

        crs_text = mapper_dict.get("crs", None)

        try:
            mesh_path = Path(mapper_dict["mesh_filepath"])
        except KeyError as err:
            raise KeyError('"mesh_filepath" key is missing') from err

        coordinate_shift = None
        if mapper_dict.get("coordinate_shift", None) is not None:
            coordinate_shift = np.array(mapper_dict["coordinate_shift"])

        crs = None

        # Get CRS if specified
        if crs_text is not None:
            try:
                crs = CRS(crs_text)
            except pyproj.exceptions.CRSError as err:
                raise CRSInputError("No valid CRS wkt string supported") from err

        mapping = cls(mesh_path=mesh_path, crs=crs, coordinate_shift=coordinate_shift)

        return mapping

    @property
    def param_dict(self):
        """Return the mapper configuration dictionary.

        :return: Mapper configuration dictionary.
        :rtype: dict[str, Any]
        """

        dict_return = {
            "type": self.type.fullname,
            "mesh_filepath": self.path.as_posix(),
            "crs": self.crs_wkt,
            "coordinate_shift": self.translation.tolist() if self.translation is not None else None,
        }
        return dict_return

    def map_coordinates_from_rays(
        self,
        ray_vectors_crs_s: ArrayNx3,
        ray_start_crs_s: ArrayNx3,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map coordinates by intersecting rays with the mesh surface.

        :param ray_vectors_crs_s: Ray direction vectors (N×3).
        :type ray_vectors_crs_s: ArrayNx3
        :param ray_start_crs_s: Ray start points (N×3), same shape as ``ray_vectors_crs_s``.
        :type ray_start_crs_s: ArrayNx3
        :param crs_s: CRS of the input rays, defaults to None.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to None.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing intersection coordinates and a validity mask.
        :rtype: MappingResult
        :raises ValueError: If input arrays have incompatible shapes.
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        rays_vector = to_array_nx3(ray_vectors_crs_s)
        ray_start = to_array_nx3(ray_start_crs_s)

        if ray_start.shape != rays_vector.shape:
            raise ValueError("Array size for ray vectors and ray start points do not fit")

        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        if coo_trafo is not None:
            dists = np.array([0, 50, 200, 700])
            points_batch = ray_start[:, None, :] + dists[None, :, None] * rays_vector[:, None, :]
            # ray_start_crs_m = coo_trafo.transform(ray_start)
            _ray_pts_crs_m = coo_trafo.transform(points_batch)
            _, rays_vector_crs_m = rays_from_points_batch(_ray_pts_crs_m, use_first_segment=True)
            ray_start_crs_m = _ray_pts_crs_m[:, 0, :]
        else:
            ray_start_crs_m = ray_start * 1.0
            rays_vector_crs_m = rays_vector * 1.0

        # Trimesh to find intersection locations
        # It returns the intersection location, the index of the ray and the index of the triangle of the mesh
        assert isinstance(self._mesh, trimesh.Trimesh)
        loc_crs_m_raw, index_ray, _ = self._mesh.ray.intersects_location(  # pyright: ignore[reportUnknownMemberType]
            ray_origins=ray_start_crs_m,
            ray_directions=rays_vector_crs_m,
            multiple_hits=False,
        )

        loc_crs_m_raw: ArrayNx3 = np.array(loc_crs_m_raw)
        if loc_crs_m_raw.size == 0:
            issue = {Issue.NO_INTERSECTION}
            return ResultFailure(ok=False, error="No intersections where found", issues=issue)

        # if index_ray.shape[0] != ray_start_crs_m.shape[0]:
        #    raise MappingError("Not for all rays, intersections were found")
        mask = np.zeros(ray_vectors_crs_s.shape[0], dtype=bool)
        mask[index_ray] = True

        issue = set()
        if not np.all(mask):
            issue = {Issue.NO_INTERSECTION}

        # Sort the resulting location array to have same order as input
        # Trimesh returns unordered array for locations - but index_ray is given
        # sort_index = np.argsort(index_ray)
        # locations_crs_m = loc_crs_m_raw[sort_index, :]

        # Transform back to source crs.
        if coo_trafo is not None:
            coo_crs_source = coo_trafo.transform(loc_crs_m_raw, direction="inverse")
        else:
            coo_crs_source = loc_crs_m_raw * 1.0

        coo_crs_result = np.empty(rays_vector.shape, dtype=float)
        coo_crs_result.fill(np.nan)
        coo_crs_result[index_ray, :] = coo_crs_source

        return MappingResultSuccess(ok=True, coordinates=coo_crs_result, mask=mask, crs=crs_s, issues=issue)

    def map_heights_from_coordinates(
        self,
        coordinates_crs_s: ArrayNx3 | ArrayNx2,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Sample heights by casting vertical rays against the mesh.

        :param coordinates_crs_s: Coordinates to sample (N×2 or N×3).
        :type coordinates_crs_s: ArrayNx3 | ArrayNx2
        :param crs_s: CRS of the input coordinates, defaults to None.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to None.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing sampled coordinates and a validity mask.
        :rtype: MappingResult
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        # originally I wanted to check for vertical ray in both direction (up, down)
        # and use the higher mesh intersection for both rays, but it could be that there are multiple hits down and up
        # Now I just change the coordinates to have a z component of 9000 as a temporary solution
        # TODO maybe make that vertical grapping more advanced see comment before

        coordinates_crs_s = to_array_nx3(coordinates_crs_s, 0.0)

        h = 9000.0
        coordinates_crs_s[:, 2] = h

        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        if coo_trafo is not None:
            coordinates_mapper_crs = coo_trafo.transform(coordinates_crs_s)
        else:
            coordinates_mapper_crs = coordinates_crs_s * 1.0

        rays_vertical = np.zeros(coordinates_crs_s.shape)
        rays_vertical[:, 2] = -1

        result_mapper_crs = self.map_coordinates_from_rays(
            ray_vectors_crs_s=rays_vertical,
            ray_start_crs_s=coordinates_mapper_crs,
            crs_s=self._crs,
        )

        if result_mapper_crs.ok is False:
            return ResultFailure(ok=False, error=result_mapper_crs.error, issues=result_mapper_crs.issues)

        # Transform back to source crs.
        # Non valid coordinates will be nan
        coo_crs_source = np.empty(coordinates_mapper_crs.shape, dtype=float)
        coo_crs_source.fill(np.nan)
        index = np.flatnonzero(result_mapper_crs.mask)
        if coo_trafo is not None:
            coo_crs_source[index, :] = coo_trafo.transform(result_mapper_crs.coordinates[index, :], direction="inverse")
        else:
            coo_crs_source[index, :] = result_mapper_crs.coordinates[index, :] * 1.0

        return MappingResultSuccess(
            ok=True, coordinates=coo_crs_source, mask=result_mapper_crs.mask, crs=crs_s, issues=result_mapper_crs.issues
        )



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Geometry helper functions used across weitsicht."""

import logging

from weitsicht.geometry.coplanar_collinear import is_coplanar
from weitsicht.geometry.interpolation_bilinear import bilinear_interpolation
from weitsicht.geometry.intersection_bilinear import multilinear_poly_intersection
from weitsicht.geometry.intersection_plane import intersection_plane, intersection_plane_mat_operation
from weitsicht.geometry.line_grid_intersection import line_grid_intersection_points, raster_index_p1_p2, vector_projection

__all__ = [
    "intersection_plane",
    "intersection_plane_mat_operation",
    "vector_projection",
    "raster_index_p1_p2",
    "line_grid_intersection_points",
    "multilinear_poly_intersection",
    "bilinear_interpolation",
    "bilinear_interpolation",
    "is_coplanar",
]

logger = logging.getLogger(__name__)



from weitsicht.mapping.base_class import MappingBase


def test_crs():
    mapper = MappingBase()

    assert mapper.crs is None


def test_crs_wkt():
    mapper = MappingBase()
    assert mapper.crs_wkt is None



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Ray/line intersections with planes."""

import numpy as np

from weitsicht.utils import ArrayNx3, MaskN_, Vector3D

__all__ = ["intersection_plane", "intersection_plane_mat_operation"]


def intersection_plane(
    line_vec: np.ndarray,
    line_point: np.ndarray,
    plane_point: np.ndarray,
    plane_normal: np.ndarray | None = None,
):
    """Intersect a line with a plane.

    If ``plane_normal`` is ``None``, the plane normal defaults to ``(0, 0, 1)``.

    :param line_vec: Direction vector of the line.
    :type line_vec: numpy.ndarray
    :param line_point: A point on the line.
    :type line_point: numpy.ndarray
    :param plane_point: A point on the plane.
    :type plane_point: numpy.ndarray
    :param plane_normal: Plane normal vector, defaults to ``None`` (interpreted as ``(0, 0, 1)``).
    :type plane_normal: numpy.ndarray | None
    :return: Tuple ``(intersection_point, direction_valid)``. If no valid intersection exists,
        the point contains NaNs and ``direction_valid`` is ``False``.
    :rtype: tuple[numpy.ndarray, bool]
    """
    epsilon = 1e-6
    # Define plane
    if plane_normal is None:
        plane_normal = np.array([0, 0, 1])

    n_dotu = plane_normal.dot(line_vec)

    # Test for parallelity or vector insi
    if abs(n_dotu) < epsilon:
        direction_valid = False
        return np.array([np.nan, np.nan, np.nan]), direction_valid
        # print "no intersection or line is within plane"

    w = line_point - plane_point
    si = -plane_normal.dot(w) / n_dotu
    p_si = w + si * line_vec + plane_point

    v1 = line_point - p_si
    v2 = line_point - (line_point + line_vec)
    cos_a = v1.dot(v2) / np.linalg.norm(v1) / np.linalg.norm(v2)

    if cos_a < 0.9:
        direction_valid = False
    else:
        direction_valid = True

    return p_si, direction_valid


def intersection_plane_mat_operation(
    line_vec: ArrayNx3,
    line_point: ArrayNx3,
    plane_point: Vector3D,
    plane_normal: Vector3D | None = None,
) -> tuple[ArrayNx3, MaskN_]:
    """Intersect multiple rays with a plane (vectorized).

    For every ray defined by ``line_point[i] + t * line_vec[i]`` this function computes the
    intersection with a plane defined by ``plane_point`` and ``plane_normal``.

    If ``plane_normal`` is ``None``, the plane normal defaults to ``(0, 0, 1)``.

    :param line_vec: Ray direction vectors.
    :type line_vec: ArrayNx3
    :param line_point: Ray origin points.
    :type line_point: ArrayNx3
    :param plane_point: A point on the plane.
    :type plane_point: Vector3D
    :param plane_normal: Plane normal vector, defaults to ``None`` (interpreted as ``(0, 0, 1)``).
    :type plane_normal: Vector3D | None
    :return: Tuple ``(intersection_points, valid_mask)``. Invalid intersections contain NaNs.
    :rtype: tuple[ArrayNx3, ``MaskN_``]
    """
    valid: np.ndarray
    valid = np.zeros((line_vec.shape[0],), dtype=bool)

    epsilon = 1e-7
    # Define plane
    if plane_normal is None:
        pln = np.array([0.0, 0.0, 1.0])
    else:
        pln = plane_normal
    # Define ray
    # rayDirection = numpy.array([0, -1, -1])
    # rayPoint = numpy.array([0, 0, 10])  # Any point along the ray

    n_dotu = line_vec.dot(pln)
    # np.einsum('ij,ij->i', plane_normal, line_vec)

    # check for parallel lines to plane
    valid = abs(n_dotu) > epsilon

    w = line_point - plane_point

    si = np.zeros(n_dotu.shape[0], dtype=float)
    si[valid] = np.dot(-w[valid, :], pln) / n_dotu[valid]

    # si * line_vec
    si_m_line_vec = np.zeros(w.shape, dtype=float)
    si_m_line_vec[valid, :] = np.einsum("ij,i->ij", line_vec[valid, :], si[valid])

    # p_si = np.zeros((w.shape[0], 3), dtype=float)
    p_si = np.empty((w.shape[0], 3), dtype=float)
    p_si.fill(np.nan)
    p_si[valid, :] = w[valid, :] + si_m_line_vec[valid, :] + plane_point

    # Check if the direction would be to the other direction aka backwards
    # because that is mathematically possible on a line plane intersection
    v1 = np.zeros(w.shape, dtype=float)
    v1[valid, :] = p_si[valid, :] - line_point[valid, :]
    valid[valid] = np.logical_and(valid[valid], np.linalg.norm(v1[valid, :], axis=1) > epsilon)
    cos_a = (
        np.einsum("ij,ij->i", v1[valid, :], line_vec[valid, :])
        / np.linalg.norm(v1[valid, :], axis=1)
        / np.linalg.norm(line_vec[valid, :], axis=1)
    )

    valid[valid] = cos_a > 0.9

    return p_si, valid


if __name__ == "__main__":
    a: ArrayNx3
    a = np.array([[0, 0, 0], [0, 0, 1], [0, 0, -1], [0, 0, -1]])
    b = np.array([[0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 0]])
    c = np.array([0, 0, 0])
    d = np.array([0, 0, 1])
    intersection_plane_mat_operation(a, b, c, d)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------
"""Shared utility types, conversion helpers, and result objects.

This module defines:
- Numpy typing aliases used throughout the package (``ArrayNx2``, ``ArrayNx3``, …)
- Helper functions for normalizing array-like inputs (``to_array_nx2``, ``to_array_nx3``)
- Result dataclasses used by projection and mapping APIs
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Generic, Literal, TypeAlias, TypeVar

import numpy as np
from numpy.typing import NDArray

# from typing_extensions import Annotated
from pyproj import CRS
from pyproj.crs.crs import CompoundCRS

__all__ = [
    "MappingResultSuccess",
    "ProjectionResult",
    "ProjectionResultSuccess",
    "MappingResult",
    "MappingResultSuccess",
    "ResultFailure",
    "Issue",
    "to_array_nx2",
    "to_array_nx3",
    "MaskN_",
    "ArrayN_",
    "Array3x3",
    "ArrayNx3",
    "ArrayNx2",
    "ArrayNxN",
    "Vector2D",
    "Vector3D",
]

_N = TypeVar("_N", bound=int)
_IssueEnum = TypeVar("_IssueEnum", bound=Enum)
#
## MaskN_: TypeAlias = npt.NDArray[np.bool_]
MaskN_: TypeAlias = NDArray[np.bool_]
ArrayN_: TypeAlias = NDArray[np.float64]

Array3x3: TypeAlias = NDArray  # [tuple[Literal[3], Literal[3]], np.dtype[np.float64]]
ArrayNx3: TypeAlias = NDArray  #  [tuple[_N, Literal[3]], np.dtype[np.float64]]
ArrayNx2: TypeAlias = NDArray  #  [tuple[_N, Literal[2]], np.dtype[np.float64]]
ArrayNxN: TypeAlias = NDArray  #  [tuple[_N, _N], np.dtype[np.float64]]

Vector2D: TypeAlias = NDArray  #  [tuple[Literal[2]], np.dtype[np.float64]]
Vector3D: TypeAlias = NDArray  #  [tuple[Literal[3]], np.dtype[np.float64]]

# _N = TypeVar("_N", bound=int)

# Typed ndarray aliases with shape hints for stricter type checking
# MaskN_: TypeAlias = NDArray[np.bool_]  # shape: (*,)
# ArrayN_: TypeAlias = NDArray[np.float64]  # shape: (*,)

# Shape annotations via Annotated to avoid numpy.typing.Shape dependency
# Array3x3: TypeAlias = Annotated[NDArray[np.float64], "...x3x3"]
# ArrayNx3: TypeAlias = Annotated[NDArray[np.float64], "...xN x3"]
# ArrayNx2: TypeAlias = Annotated[NDArray[np.float64], "...xN x2"]
# ArrayNxN: TypeAlias = NDArray[np.float64]

# Vector2D: TypeAlias = Annotated[NDArray[np.float64], "len=2"]
# Vector3D: TypeAlias = Annotated[NDArray[np.float64], "len=3"]


def to_array_nx2(
    array_like_nx2: list[int] | list[list[int]] | list[float] | list[list[float]] | ArrayNx2 | ArrayNx3,
) -> ArrayNx2:
    """Convert array-like input to an N×2 numpy array.

    :param array_like_nx2: Array-like input interpreted as 2D coordinates.
    :type array_like_nx2: list[int] | list[list[int]] | list[float] | list[list[float]] | ArrayNx2 | ArrayNx3
    :return: Array of shape N×2.
    :rtype: ArrayNx2
    :raises ValueError: If the input cannot be interpreted as an N×2 array.
    """

    _array_like_nx2 = np.array(array_like_nx2)
    if _array_like_nx2.ndim == 1:
        _array_like_nx2 = np.array([_array_like_nx2])

    if _array_like_nx2.shape[1] >= 2:
        _array_like_nx2 = _array_like_nx2[:, :2]
    else:
        raise ValueError("Dimensions of 2D array not fitting")

    return _array_like_nx2


def to_array_nx3(
    array_like_nx3: list[float] | list[list[float]] | ArrayNx2 | ArrayNx3, fill_z: float | None = None
) -> ArrayNx3:
    """Convert array-like input to an N×3 numpy array.

    If the input is N×2 and ``fill_z`` is provided, a Z column is appended. Otherwise a
    :class:`ValueError` is raised for inputs that cannot be interpreted as N×3.

    :param array_like_nx3: Array-like input interpreted as 3D coordinates.
    :type array_like_nx3: list[float] | list[list[float]] | ArrayNx2 | ArrayNx3
    :param fill_z: Z value used when the input is N×2, defaults to ``None``.
    :type fill_z: float | None
    :return: Array of shape N×3.
    :rtype: ArrayNx3
    :raises ValueError: If the input cannot be interpreted as an N×3 array.
    """

    _array_like_nx3 = np.array(array_like_nx3)

    if _array_like_nx3.ndim == 1:
        _array_like_nx3 = np.array([_array_like_nx3])

    if _array_like_nx3.shape[1] >= 3:
        _array_like_nx3 = _array_like_nx3[:, :3]
    else:
        if fill_z is not None:
            _array_like_nx3 = np.hstack((_array_like_nx3, np.zeros((_array_like_nx3.shape[0], 1))))
        else:
            raise ValueError("Dimension of 3D array not fitting")

    return _array_like_nx3


class Issue(Enum):
    """Issue codes used in result objects to describe runtime outcomes."""

    # POSITION_MISSING = "position_missing"
    # ORIENTATION_MISSING = "orientation_missing"
    # CAMERA_MISSING = "camera_missing"
    # CRS_MISSING = "crs_missing"
    # MAPPER_MISSING = "mapper_missing"
    # GEO_TRANSFORM_MISSING = "Geo Transform for orthophoto missing"
    # DIMENSION_MISMATCH = "Dimensions do not match"
    WRONG_DIRECTION = "ray probably in the wrong direction"
    OUTSIDE_RASTER = "ray is outside raster"
    RASTER_NO_DATA = "touched no data raster cells"
    MAX_ITTERATION = "max itteration reached"
    NO_INTERSECTION = "No intersection was found"
    INVALID_PROJECTIIONS = "Projections are not within image border"
    IMAGE_BATCH_ERROR = "A single image in image batch calls raised an error"
    UNKNOWN = "unknown"


@dataclass
class ResultFailure(Generic[_IssueEnum]):
    """Failure result for projection and mapping operations."""

    ok: Literal[False]
    error: str
    issues: set[_IssueEnum] = field(default_factory=set)


@dataclass
class ProjectionResultSuccess:
    """Successful projection result."""

    ok: Literal[True]
    pixels: ArrayNx2
    mask: MaskN_
    issues: set[Issue] = field(default_factory=lambda: set[Issue]())


ProjectionResult = ProjectionResultSuccess | ResultFailure[Issue]


@dataclass
class MappingResultSuccess:
    """Successful mapping result."""

    ok: Literal[True]
    coordinates: ArrayNx3
    mask: MaskN_  # mapped 3D coords
    crs: CRS | CompoundCRS | None = None
    gsd: float | None = None  # mean GSD (or scalar for center)
    gsd_per_point: ArrayN_ | None = None  # optional per-point GSDs
    area: float | None = None
    issues: set[Issue] = field(default_factory=lambda: set[Issue]())


# @dataclass
# class MappingResult:
#    ok: bool
#    coordinates: Optional[np.ndarray] = None
#    crs: Optional[CRS | CompoundCRS | None] = None
#    mask: Optional[np.ndarray] = None   # mapped 3D coords
#    gsd: Optional[float] = None           # mean GSD (or scalar for center)
#    gsd_per_point: Optional[np.ndarray] = None  # optional per-point GSDs
#    area: Optional[float] = None          # footprint area (if applicable)
#    error: Optional[list[str]] = None
#    issues: set[Issue] = field(default_factory=set)   # e.g. ["position", "orientation"]


MappingResult = MappingResultSuccess | ResultFailure[Issue]
